# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import numpy as np
from . import CNNSizeFactor


class ModelHyperparams(object):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, hprm: dict | None = None):
    self.hprm = hprm
    self.model_name = None
    self.base_name = ""
    self.architecture = ""
    self.variant = ""
    self.class_count = 0
    self.input_dims = []
    self.dropout_prob = None
    # self.kind = "" //TODO: Validate expected input according to model kind
  
  # --------------------------------------------------------------------------------------------------------------------
  @property
  def input_channels(self):
    '''
    :return: The channels from the dimensions of an image input sample
    '''
    return self.input_dims[-1]
  
  # --------------------------------------------------------------------------------------------------------------------
  @property
  def input_resolution(self):
    '''
    :return: The resolution from the dimensions of an image input sample
    '''
    return self.input_dims[:2]
  
  # --------------------------------------------------------------------------------------------------------------------
  @property
  def cnn_size_factor(self):
    '''
      Calculates the cnn size factor using the maximum dimension as log normalized where 0 is the size for MNIST input
      and 1 the size for ImageNet input.
    :return: 1  tiny images (MNIST, FMNIST, CIFAR10, CIFAR100)
             2  small images (ImageNet 64x64)
             3
             4
    '''
    
    def get_id(x):
      if 0.0 <= x < 0.1:
        return 1
      elif 0.1 <= x < 0.4:
        return 2
      elif 0.4 <= x < 0.8:
        return 3
      elif 0.8 <= x <= 1.0:
        return 4
      else:
        return None
    
    nHeight, nWidth = self.input_dims[:2]
    nDim = nHeight
    if nHeight < nWidth:
      Dim = nWidth
    nLogDim = np.log10(nDim)
    nMin = np.log10(28)
    nMax = np.log10(300)
    nLogNormalized = (nLogDim - nMin) / (nMax - nMin)
    nSizeFactor = CNNSizeFactor(get_id(nLogNormalized))
    return nSizeFactor
  
  # --------------------------------------------------------------------------------------------------------------------
  @classmethod
  def from_dict(cls, hyperparams):
    '''
    Deserializes a hyperparameters object from a dictionary
    
    :param hyperparams: A dictionary containing the architectural hyperparameters for the model
    :return: a new model hyperparams object
    '''
    oModelHyperparams = ModelHyperparams(hyperparams)
    oModelHyperparams.model_name = hyperparams.get("Model.Name", "?")
    
    oModelHyperparams.base_name = oModelHyperparams.model_name
    sParts = oModelHyperparams.base_name.split(".")
    oModelHyperparams.architecture = "base"
    if len(sParts) >= 2:
      oModelHyperparams.base_name = sParts[0]
      oModelHyperparams.architecture = sParts[1].lower()
    
    oModelHyperparams.variant = hyperparams.get("Model.Variant", "")
    oModelHyperparams.name = hyperparams.get("Model.Name", "?")
    oModelHyperparams.input_dims = hyperparams.get("Dataset.SampleDims", [])
    oModelHyperparams.class_count = hyperparams.get("Dataset.ClassCount", 0)
    nDropOutKeepProb = hyperparams.get("Training.Regularize.DropOut.KeepProb", None)
    if nDropOutKeepProb is None:
      oModelHyperparams.dropout_prob = hyperparams.get("Training.Regularize.DropOut.DropProb", 0.0)
    else:
      oModelHyperparams.dropout_prob = 1.0 - nDropOutKeepProb
    
    return oModelHyperparams
  # --------------------------------------------------------------------------------------------------------------------



from .subset_type import SubsetType

class DataSetFolder(object):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, folder_name, filestore):
    self.folder_name = folder_name
    self.filestore = filestore
    self.filestore_ts = None
    self.filestore_vs = None
    self.filestore_ut = None
    self.split_filestores = []

    self.subfolders = self.filestore.list_folders(is_full_path=False)
    self.is_split, sTSFolder, sVSFolder, sUTFolder = self.get_split_subfolders(self.subfolders)
    if self.is_split:
      if sTSFolder is not None:
        self.filestore_ts = self.filestore.subfs(sTSFolder, must_exist=True)
        self.split_filestores.append(self.filestore_ts)
      if sVSFolder is not None:
        self.filestore_vs = self.filestore.subfs(sVSFolder, must_exist=True)
        self.split_filestores.append(self.filestore_vs)
      if sUTFolder is not None:
        self.filestore_ut = self.filestore.subfs(sUTFolder, must_exist=True)
        self.split_filestores.append(self.filestore_ut)
  # --------------------------------------------------------------------------------------------------------------------
  def get_split_subfolders(self, folders):
    sTSFolder = None
    sVSFolder = None
    sUTFolder = None
    bIsSplit = False
    for sFolder in folders:
      oFolderSubsetType = SubsetType(sFolder)
      if oFolderSubsetType.is_training_set:
        sTSFolder = sFolder
        bIsSplit = True
      elif oFolderSubsetType.is_validation_set:
        sVSFolder = sFolder
        bIsSplit = True
      elif oFolderSubsetType.is_unknown_test_set:
        sUTFolder = sFolder
        bIsSplit = True

    #  When confusing terminology is uses and 'test' means 'validation'
    if (sUTFolder is not None) and (sVSFolder is None):
      sVSFolder = sUTFolder
      sUTFolder = None

    return bIsSplit, sTSFolder, sVSFolder, sUTFolder
  # --------------------------------------------------------------------------------------------------------------------
  def __str__(self):
    return "./" + self.folder_name
  # --------------------------------------------------------------------------------------------------------------------
  def __repr__(self):
    return self.__str__()
  # --------------------------------------------------------------------------------------------------------------------


from enum import Enum

class MLTask(Enum):
  BINARY_CLASSIFICATION = 0
  MULTICLASS_CLASSIFICATION = 1
  REGRESSION = 2
  MULTIVARIATE_REGRESSION = 3
  CLUSTERING = 4
  INPUT_RECONSTRUCTION = 5
  RANKING = 6


def get_task_from_hyperparams(hyperparams: dict) -> MLTask:
  sTask = None
  if "Experiment.Task" in hyperparams:
    sTask = hyperparams["Experiment.Task"]
    
  oTask = None
  if sTask is not None:
    sTask = sTask.upper().replace(" ", "_")
    if sTask in MLTask.__members__:
      oTask = MLTask[sTask]
      
  return oTask


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
import numpy as np
import locale
from .fileobject import FileObject


class TextFile(FileObject):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None, is_verbose=False):
    super(TextFile, self).__init__(filename, parent_folder, error_template)
    self.is_verbose = is_verbose
    self._opened_filename = None
    self._encoding = None
  # ----------------------------------------------------------------------------------
  def open(self, filename, encoding=None):
    self._opened_filename = filename
    self._encoding = encoding
    return self

  # ----------------------------------------------------------------------------------
  def close(self):
    self._opened_filename = None
    self._encoding = None
  # ----------------------------------------------------------------------------------
  @property
  def rows(self):
    sFileName = self._useFileName(self._opened_filename)

    if os.path.isfile(sFileName):
      if self._encoding is not None:
        oEncodingToTry = [self._encoding]
      else:
        oEncodingToTry = ["utf-8", "utf-16", "latin1", "ascii"]  # Add more if needed
      bIsLoaded = False
      for sEnc in oEncodingToTry:
        try:
          with open(sFileName, "r", encoding=sEnc) as oFile:
            yield oFile.read()
          bIsLoaded = True
          break
        except (UnicodeDecodeError, UnicodeError):
          continue
      if not bIsLoaded:
        raise ValueError("Unsupported encoding")
  # ----------------------------------------------------------------------------------
  def load(self, filename=None, encoding=None):
    filename = self._useFileName(filename)

    sText = None
    if os.path.isfile(filename):
      if self._encoding is not None:
        oEncodingToTry = [encoding]
      else:
        oEncodingToTry = ["utf-8", "utf-16", "latin1", "ascii"]  # Add more if needed

      bIsLoaded = False
      for sEnc in oEncodingToTry:
        try:
          with open(filename, "r", encoding=sEnc) as oFile:
            sText = oFile.read()
          bIsLoaded = True
          break
        except (UnicodeDecodeError, UnicodeError):
          continue
      if not bIsLoaded:
        raise ValueError("Unsupported encoding")

    return sText
  # --------------------------------------------------------------------------------------------------------------------
  def save(self, text_obj, filename=None, encoding="utf-8"):
    sFilename = self._useFileName(filename)

    """
    Writes text to a file

    Parameters
        p_sFileName        : Full path to the text file
        p_sText            : Text to write
    """
    if self.is_verbose:
      print("  {.} Saving text to %s" % sFilename)

    bIsIterable = False
    if isinstance(text_obj, list):
      bIsIterable = True
    if isinstance(text_obj, np.ndarray):
      bIsIterable = text_obj.dtype = str

    if bIsIterable:
      with open(sFilename, "w", encoding=encoding) as oFile:
        for sLine in text_obj:
          print(sLine, file=oFile)
        oFile.close()
    else:
      with open(sFilename, "w", encoding=encoding) as oFile:
        print(text_obj, file=oFile)
        oFile.close()
    return True
  # --------------------------------------------------------------------------------------------------------------------


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................


class DataHyperparams(object):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, hprm: dict | None = None):
    self.hprm = hprm
    self.dataset_name = None
    self.dataset_fs_name = None
    self.fold_number = None
    self.class_count = 0
    self.sample_dims = None
  # --------------------------------------------------------------------------------------------------------------------
  @classmethod
  def from_dict(cls, hyperparams):
    '''
    Deserializes a hyperparameters object from a dictionary

    :param hyperparams: A dictionary containing the architectural hyperparameters for the model
    :return: a new model hyperparams object
    '''
    oModelHyperparams = DataHyperparams(hyperparams)
    oModelHyperparams.dataset_name = hyperparams.get("Dataset.Name", "?")
    oModelHyperparams.dataset_fs_name = oModelHyperparams.dataset_name.lower()
    oModelHyperparams.fold_number = hyperparams.get("Dataset.FoldNumber", 1)
    oModelHyperparams.class_count = hyperparams.get("Dataset.ClassCount", 0)
    oModelHyperparams.sample_dims = hyperparams.get("Dataset.SampleDims", 0)
    
    return oModelHyperparams
  # --------------------------------------------------------------------------------------------------------------------




from .dataset_base import DataSetBase
from .image_dataset import ImageDataSet
from .sample_set import SampleSet
from .data_feed import DataFeed
from .subset_type import SubsetType
from .sample_set import SampleSet
from radnn import mlsys
if mlsys.is_tensorflow_installed:
  from .tf_classification_data_feed import TFClassificationDataFeed

from .image_dataset_files import ImageDataSetFiles




# ......................................................................................
# MIT License

# Copyright (c) 2020-2022 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................


import numpy as np
import matplotlib.pyplot as plt  # use the subpackage (a.k.a. namespace) with the alias "plt"
from matplotlib import colors

# =========================================================================================================================
class PlotHistogramOfClasses(object):  # class CPlot: object
  # --------------------------------------------------------------------------------------
  def __init__(self, data, classes, is_probabilities=False):
    self.data = data
    self.classes = classes
    self.is_probabilities = is_probabilities
  
  # --------------------------------------------------------------------------------------
  def prepare(self):
    
    fig, ax = plt.subplots(figsize=(7, 7))
    
    ax.hist(self.data, density=self.is_probabilities, bins=self.classes, ec="k")
    ax.locator_params(axis='x', integer=True)
    
    if self.is_probabilities:
      plt.ylabel('Probabilities')
    else:
      plt.ylabel('Counts')
    plt.xlabel('Classes')
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def save(self, filename):
    plt.savefig(filename, bbox_inches='tight')
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def show(self):
    plt.show()
  # --------------------------------------------------------------------------------------------------------------------
  


# =========================================================================================================================




import numpy as np
from radnn.core import RequiredLibs
oReqs = RequiredLibs()
if oReqs.is_opencv_installed:
  import cv2
from PIL import Image, ImageChops, ImageFilter, ImageDraw
from .colors import color

phi=(1.0+np.sqrt(5.0))/2.0

class ImageProcessor(object):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, pixels=None, filename=None, image=None, bgcolor=None):
    self.pixels = pixels
    self.filename = filename
    self.image = image
    self.bgcolor = bgcolor
    self._size = None

    if self.pixels is not None:
      if self.pixels is not None:
        nShape = self.pixels.shape
        if len(nShape) == 4:
          self._size = nShape[1:4]
        elif len(nShape) == 3:
          self._size = nShape[1:3]
        elif len(nShape) == 2:
          self._size = nShape

    if image is not None:
      self.pixels = np.array(self.image)
  # --------------------------------------------------------------------------------------------------------------------
  @property
  def size(self):
    if self.image is not None:
      return self.image.size
    else:
      return self._size
  # --------------------------------------------------------------------------------------------------------------------
  def load(self, filename=None):
    if filename is not None:
      self.filename = filename
    self.image = Image.open(self.filename).convert("RGB")
    self.pixels = np.array(self.image)
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def pad_square_with_edges(self, size=(227, 227)):
    img = self.image
    img_width, img_height = img.size

    # Determine the scaling factor to maintain aspect ratio
    if img_width > img_height:
      bIsLandscape = True
      ratio = img_height / img_width
      new_width = size[0]
      new_height = int(size[0] * ratio)
    else:
      bIsLandscape = False
      ratio = img_width / img_height
      new_width = int(size[1] * ratio)
      new_height = size[1]

    # Resize the image while maintaining aspect ratio
    img = img.resize((new_width, new_height), Image.LANCZOS)

    # Create a blank canvas for the final image
    Result = np.zeros((size[1], size[0], 3), dtype=np.uint8)
    img_array = np.array(img)

    # Center the resized image on the new canvas
    offset_x = (size[0] - new_width) // 2
    offset_y = (size[1] - new_height) // 2
    Result[offset_y:offset_y + new_height, offset_x:offset_x + new_width, :] = img_array

    # Fill edges with repeated stripes and blur for a smooth effect
    if bIsLandscape:
      # Top and bottom padding
      first_row = Result[offset_y, :, :]
      last_row = Result[offset_y + new_height - 1, :, :]

      top_rows = np.repeat(first_row.reshape(1, size[0], 3), offset_y, axis=0)
      bottom_rows = np.repeat(last_row.reshape(1, size[0], 3), size[1] - offset_y - new_height, axis=0)

      # Apply blur to soften edges
      im_top = Image.fromarray(top_rows).filter(ImageFilter.BLUR)
      im_bottom = Image.fromarray(bottom_rows).filter(ImageFilter.BLUR)

      Result[0:offset_y, :, :] = np.array(im_top)
      Result[offset_y + new_height:size[1], :, :] = np.array(im_bottom)
    else:
      # Left and right padding
      first_col = Result[:, offset_x, :]
      last_col = Result[:, offset_x + new_width - 1, :]

      left_cols = np.repeat(first_col.reshape(size[1], 1, 3), offset_x, axis=1)
      right_cols = np.repeat(last_col.reshape(size[1], 1, 3), size[0] - offset_x - new_width, axis=1)

      # Apply blur to soften edges
      im_left = Image.fromarray(left_cols).filter(ImageFilter.BLUR)
      im_right = Image.fromarray(right_cols).filter(ImageFilter.BLUR)

      Result[:, 0:offset_x, :] = np.array(im_left)
      Result[:, offset_x + new_width:size[0], :] = np.array(im_right)

    # Convert back to a PIL image
    final_img = Image.fromarray(Result)
    return ImageProcessor(image=final_img, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def _get_background_color(self, bgcolor):
    if bgcolor is not None:
      self.bgcolor = bgcolor
    if self.bgcolor is None:
      self.bgcolor = "black"
    if isinstance(self.bgcolor, str):
      nBGColor = color(self.bgcolor)
    else:
      nBGColor = self.bgcolor
    return nBGColor
  # --------------------------------------------------------------------------------------------------------------------
  def roll(self, shift_x=0, shift_y=0, bgcolor=None):
    nBGColor = self._get_background_color(bgcolor)
    img = self.image

    arr = np.array(img)

    # Apply roll effect
    arr = np.roll(arr, shift_x, axis=1)  # Shift horizontally
    arr = np.roll(arr, shift_y, axis=0)  # Shift vertically

    img_out = Image.fromarray(arr)


    return ImageProcessor(image=img_out, filename=self.filename, bgcolor=self.bgcolor)

  # --------------------------------------------------------------------------------------------------------------------
  def zoom_center(self, scale=1.2, bgcolor=None):
    nBGColor = self._get_background_color(bgcolor)
    img = self.pixels
    w, h = self.size

    tx, ty = 0, 0
    if scale < 1.0:
      tx = int((w * (1.0 - scale)) // 2)
      ty = int((h * (1.0 - scale)) // 2)
    else:
      tx = - int(w*(scale - 1.0) // 2)
      ty = - int(w*(scale - 1.0) // 2)

    # Transformation matrix for scaling and translation
    M = np.array([[scale, 0, tx],
                  [0, scale, ty]], dtype=np.float32)

    # Apply affine warp
    result = cv2.warpAffine(img, M, (w, h))

    oImage = Image.fromarray(result)
    oCropped = oImage.crop((tx, ty, w - tx - 1, h - ty - 1)).convert("RGBA")
    oNewImageWithBackground = Image.new("RGBA",self.size, nBGColor)
    oNewImageWithBackground.paste(oCropped, (tx, ty), oCropped)

    return ImageProcessor(image=oNewImageWithBackground, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def zoom_pan(self, scale=1.2, tx=20, ty=20, bgcolor=None):
    nBGColor = self._get_background_color(bgcolor)
    img = self.pixels
    w, h = self.size

    # Transformation matrix for scaling and translation
    M = np.array([[scale, 0, tx],
                  [0, scale, ty]], dtype=np.float32)

    # Apply affine warp
    result = cv2.warpAffine(img, M, (w, h))

    oImage = Image.fromarray(result)
    oCropped = oImage.crop((tx, ty, w*scale - tx, h*scale - ty)).convert("RGBA")
    oNewImageWithBackground = Image.new("RGBA",self.size, nBGColor)
    oNewImageWithBackground.paste(oCropped, (tx, ty), oCropped)

    return ImageProcessor(image=oNewImageWithBackground, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def wave_effect(self, amplitude=20, frequency=0.1):
    img = self.pixels
    w, h = self.size

    # Create mapping arrays
    map_x = np.zeros((h, w), dtype=np.float32)
    map_y = np.zeros((h, w), dtype=np.float32)

    for i in range(h):
      for j in range(w):
        offset_x = int(amplitude * np.sin(2 * np.pi * frequency * i))
        map_x[i, j] = j + offset_x
        map_y[i, j] = i

    # Apply remapping
    result = cv2.remap(img, map_x, map_y, cv2.INTER_LINEAR)
    pil_image = Image.fromarray(result)
    return ImageProcessor(image=pil_image, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def rotate(self, degrees, is_original_scale=True, bgcolor=None):
    nBGColor = self._get_background_color(bgcolor)
    oImageRGBA = self.image.convert('RGBA')
    oImageRotated = oImageRGBA.rotate(degrees, expand=True)
    oBackground = Image.new('RGBA', oImageRotated.size, nBGColor)
    oNewImage = Image.composite(oImageRotated, oBackground, oImageRotated).convert(self.image.mode)

    if is_original_scale:
      orig_w, orig_h = self.image.size
      new_w, new_h = oNewImage.size
      left = (new_w - orig_w) // 2
      top = (new_h - orig_h) // 2
      right = left + orig_w
      bottom = top + orig_h

      # Crop to original aspect ratio
      oNewImage = oNewImage.crop((left, top, right, bottom))
    else:
      # Composite the rotated image onto the background using its alpha mask
      oNewImage = Image.composite(oImageRotated, oBackground, oImageRotated).convert(self.image.mode)

    return ImageProcessor(image=oNewImage, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def fit_to_size(self, size=(227, 227), bgcolor=None):
    nBGColor = self._get_background_color(bgcolor)

    img = self.image
    img_width = float(self.size[0])
    img_height = float(self.size[1])

    if img_width > img_height:
      bIsLandscape = True
      ratio = img_height / img_width
      new_width = size[0]
      new_height = int(size[0] * ratio)
    else:
      bIsLandscape = False
      ratio = img_width / img_height
      new_width = int(size[0] * ratio)
      new_height = size[0]

    img = img.resize((new_width, new_height), Image.NONE).convert("RGBA")

    nOffsetX = 0
    nOffsetY = 0
    if bIsLandscape:
      nOffsetY = (size[1] - img.size[1]) // 2
    else:
      nOffsetX = (size[0] - img.size[0]) // 2

    thumb = img.crop((0, 0, size[0], size[1]))

    oMovedImage = ImageChops.offset(thumb, int(nOffsetX), int(nOffsetY))

    oNewImageWithBackground = Image.new("RGBA", oMovedImage.size, nBGColor)
    oNewImageWithBackground.paste(oMovedImage, (0, 0), oMovedImage)

    return ImageProcessor(image=oNewImageWithBackground, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def crop_to_size(self, target_size=(227, 227)):
    """
    Resizes the image so that the smallest dimension matches the corresponding
    target size dimension, then center crops it to the exact target size.

    :param image: PIL Image to be processed.
    :param target_size: Tuple (width, height) of the desired output size.
    :return: Cropped PIL Image of the specified size.
    """
    target_w, target_h = target_size
    img_w, img_h = self.size

    # Determine scale factor to match the smallest dimension
    scale = max(target_w / img_w, target_h / img_h)
    new_size = (int(img_w * scale), int(img_h * scale))

    # Resize while maintaining aspect ratio
    oResizedImage = self.image.resize(new_size, Image.NONE) #Image.LANCZOS

    # Center crop to the exact target size
    left = (oResizedImage.width - target_w) / 2
    top = (oResizedImage.height - target_h) / 2
    right = left + target_w
    bottom = top + target_h

    oCroppedImage = oResizedImage.crop((left, top, right, bottom))

    return ImageProcessor(image=oCroppedImage, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def horizontal_wave_effect(self, amplitude=7, frequency=0.01, bgcolor=None):
    nBGColor = self._get_background_color(bgcolor)

    img = self.pixels
    w, h = self.size

    # Create mapping arrays
    map_x = np.zeros((h, w), dtype=np.float32)
    map_y = np.zeros((h, w), dtype=np.float32)

    for i in range(h):
      for j in range(w):
        offset_y = int(amplitude * np.sin(2 * np.pi * frequency * j))  # Horizontal wave
        map_x[i, j] = j
        map_y[i, j] = i + offset_y  # Apply wave effect in Y direction

    # Apply remapping
    result = cv2.remap(img, map_x, map_y, cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT, borderValue=nBGColor)
    oImage = Image.fromarray(result)
    return ImageProcessor(image=oImage, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def drop_shadow(self, shrink=0.9, shadow_offset=(8, 8), shadow_blur=15,
                            shadow_color=(32, 32, 32, 255), bgcolor=None):
    """
    Shrinks an image while keeping the final output dimensions the same by adding a drop shadow.

    Parameters:
        shrink_factor (float): The factor by which the image is shrunk (0 < shrink_factor ≤ 1).
        shadow_offset (tuple): The (x, y) offset for the shadow.
        shadow_blur (int): The blur radius for the shadow.
        shadow_color (tuple): RGBA color of the shadow (default is semi-transparent black).

    Returns:
        ImageProcessor: A new instance with the processed image.
    """
    nBGColor = self._get_background_color(bgcolor)
    assert shrink < 1.0, "The shrink factor should be less than 1.0"
    img = self.pixels
    w, h = self.size

    # Compute new size for the shrunk image
    new_w = int(w * shrink)
    new_h = int(h * shrink)

    # Place the shrunk image centered in the original dimensions
    x_center = (w - new_w) // 2
    y_center = (h - new_h) // 2


    # Resize the image (shrink)
    shrunk_img = cv2.resize(img, (new_w, new_h), interpolation=cv2.INTER_AREA)

    # Create transparent background
    shadow_img = np.zeros((h, w, 4), dtype=np.uint8)
    shadow_img[:,:,:4] = list(nBGColor)[:]

    # Create shadow by filling an ellipse or rectangle (based on image shape)
    shadow = np.full((new_h, new_w, 4), shadow_color, dtype=np.uint8)

    # Blur the shadow
    shadow = cv2.GaussianBlur(shadow, (shadow_blur, shadow_blur), 0)

    # Position the shadow
    x_offset, y_offset = x_center + shadow_offset[0], y_center + shadow_offset[1]
    shadow_img[y_offset:y_offset + new_h, x_offset:x_offset + new_w] = shadow

    # Convert shrunk image to 4-channel RGBA if not already
    if shrunk_img.shape[2] == 3:
      shrunk_img = cv2.cvtColor(shrunk_img, cv2.COLOR_RGB2RGBA)

    # Paste the shrunk image onto the shadow
    shadow_img[y_center:y_center + new_h, x_center:x_center + new_w] = shrunk_img

    # Convert back to PIL image
    oImage = Image.fromarray(shadow_img)

    return ImageProcessor(image=oImage, filename=self.filename, bgcolor=self.bgcolor)
  # --------------------------------------------------------------------------------------------------------------------
  def make_mbsf_augmented_square(self, size=(227, 227)):
    '''
    Used in MSBF paper
    :param size:
    :return:
    '''
    nSize = size

    nHalfSize = (nSize[0] // 2, nSize[1] // 2)
    nModX = nSize[0] % 2
    nModY = nSize[1] % 2

    img = self.image

    img_width = float(self.size[0])
    img_height = float(self.size[1])
    nAspectRatio = img_width / img_height
    if nAspectRatio > 1.0:
      bIrregular = nAspectRatio > (phi * 0.9)
      bIsTopBottomPadding = True
    else:
      bIrregular = nAspectRatio < (1.0 / (phi * 0.9))
      bIsTopBottomPadding = False

    # print("[%d,%d]  AspectRatio:%.4f  Irregular:%r" % (img_width, img_height, nAspectRatio, bIrregular))
    nRatioWidth = 1.0
    nRatioHeight = 1.0
    if bIrregular:
      if img_width > img_height:
        nRatioHeight = img_height / img_width
      else:
        nRatioWidth = img_width / img_height
    else:
      if img_width > img_height:
        nRatioWidth = img_width / img_height
      else:
        nRatioHeight = img_height / img_width

    new_width = int(nSize[0] * nRatioWidth)
    new_height = int(nSize[1] * nRatioHeight)

    img = img.resize((new_width, new_height), Image.NONE)
    # print("New Image Size", self.size)

    if bIrregular:
      thumb = img.crop((0, 0, nSize[0], nSize[1]))

      offset_x = int(max((nSize[0] - self.size[0]) / 2, 0))
      offset_y = int(max((nSize[1] - self.size[1]) / 2, 0))

      img = ImageChops.offset(thumb, offset_x, offset_y)

      Result = np.array(img)

      # TODO: Fadding out by number of space size
      if bIsTopBottomPadding:
        space_size_top = offset_y
        space_size_bottom = nSize[1] - new_height - offset_y
        # print("top %i, bottom %i" %(space_size_top, space_size_bottom))

        first_row = Result[offset_y + 1, :, :]
        last_row = Result[offset_y + new_height - 1, :, :]
        # first_row=np.repeat( np.mean(first_row, axis=0).reshape(1, Result.shape[2]),  Result.shape[1], axis=0)
        # last_row=np.repeat( np.mean(first_row, axis=0).reshape(1, Result.shape[2]),  Result.shape[1], axis=0 )

        top_rows = np.repeat(first_row.reshape(1, Result.shape[1], Result.shape[2]), space_size_top + 1, axis=0)
        bottom_rows = np.repeat(last_row.reshape(1, Result.shape[1], Result.shape[2]), space_size_bottom, axis=0)

        im1 = Image.fromarray(top_rows)
        im1 = im1.filter(ImageFilter.BLUR)
        top_rows = np.array(im1)

        im2 = Image.fromarray(bottom_rows)
        im2 = im2.filter(ImageFilter.BLUR)
        bottom_rows = np.array(im2)

        Result[0:offset_y + 1, :, :] = top_rows[:, :, :]
        Result[offset_y + new_height:nSize[1], :, :] = bottom_rows[:, :, :]
      else:

        space_size_left = offset_x
        space_size_right = nSize[0] - new_width - space_size_left
        # print("left %i, right %i" %(space_size_left, space_size_left))

        first_col = Result[:, offset_x + 1, :]
        last_col = Result[:, offset_x + new_width - 1, :]

        left_cols = np.repeat(first_col.reshape(Result.shape[0], 1, Result.shape[2]), space_size_left + 1, axis=1)
        right_cols = np.repeat(last_col.reshape(Result.shape[0], 1, Result.shape[2]), space_size_right, axis=1)

        im1 = Image.fromarray(left_cols)
        im1 = im1.filter(ImageFilter.BLUR)
        left_cols = np.array(im1)

        im2 = Image.fromarray(right_cols)
        im2 = im2.filter(ImageFilter.BLUR)
        right_cols = np.array(im2)

        Result[:, 0:offset_x + 1, :] = left_cols[:, :, :]
        Result[:, offset_x + new_width:nSize[0], :] = right_cols[:, :, :]

      img = Image.fromarray(Result)

      # print("Base Image Size", self.size)
    # plt.imshow(np.array(img))
    # plt.show()


    if nAspectRatio > 1.0:
      nDiff = (self.size[0] - self.size[1]) // 2
    else:
      nDiff = (self.size[1] - self.size[0]) // 2
    #
    #
    #     if False:
    #         a4im = Image.new('RGB',
    #                          (595, 842),   # A4 at 72dpi
    #                          (255, 255, 255))  # White
    #         a4im.paste(img, img.getbbox())  # Not centered, top-left corne
    #         plt.imshow(np.array(a4im))
    #         plt.show()
    nCenterX = self.size[0] // 2
    nCenterY = self.size[1] // 2

    nImgCropped = [None] * 3
    if nDiff > 40:
      nCropPositions = [0, -nDiff // 2, nDiff // 2]
    else:
      nCropPositions = [0]

    for nIndex, nShiftPos in enumerate(nCropPositions):
      nPosX = nCenterX
      nPosY = nCenterY
      if nAspectRatio > 1.0:
        nPosX += nShiftPos
      else:
        nPosY += nShiftPos

      nLeft = nPosX - nHalfSize[0]
      nRight = nPosX + nHalfSize[0] + nModX

      nTop = nPosY - nHalfSize[1]
      nBottom = nPosY + nHalfSize[1] + nModY
      nImgCropped[nIndex] = np.array(img.crop((nLeft, nTop, nRight, nBottom)))

    if len(nCropPositions) == 1:
      nImgCropped[1] = np.array(self.rotate(img, 12).pixels)
      nImgCropped[2] = np.array(self.rotate(img, -12).pixels)

    return nImgCropped


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import pandas as pd
import numpy as np

from .fileobject import FileObject

class CSVFile(FileObject):
  # ----------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None):
    super(CSVFile, self).__init__(filename, parent_folder, error_template, "txt")
  # ----------------------------------------------------------------------------------
  def load(self, filename=None, delimiter=",", encoding="utf-8"):
    sFileName = self._useFileName(filename)
    dResult = pd.read_csv(sFileName, delimiter=delimiter, encoding=encoding)
    return dResult
  # ----------------------------------------------------------------------------------
  # TODO: Iterate large csv files row by row as done in TextFile
  #@property
  #def rows(self):
  # ----------------------------------------------------------------------------------
  def save(self, obj, filename=None, headers=None, delimiter=",", encoding="utf-8"):
    sFileName = self._useFileName(filename)

    if isinstance(obj, dict):
      nMaxRowCols = 0
      oList = []
      for k, v in obj.items():
        oRow = [k]
        if isinstance(v, list):
          oRow.extend(v)
        else:
          oRow.append(v)
        if len(oRow) > nMaxRowCols:
          nMaxRowCols = len(oRow)
        oList.append(oRow)

      obj = []
      for oRow in oList:
        oRowToAdd = oRow + [None]*(nMaxRowCols - len(oRow))
        obj.append(oRowToAdd)
      if len(headers) < nMaxRowCols:
        headers = headers + [None]*(nMaxRowCols - len(headers))

    if isinstance(obj, list):
      df = pd.DataFrame(obj, columns=headers)
      df.to_csv(sFileName, index=False, header=(headers is not None), sep=delimiter, encoding=encoding)

    elif isinstance(obj, np.ndarray):
      if len(headers) < obj.shape[1]:
        headers = headers + [None]*(obj.shape[1] - len(headers))
      df = pd.DataFrame(obj, columns=headers)
      df.to_csv(sFileName, index=False, header=(headers is not None), sep=delimiter, encoding=encoding)
  # ----------------------------------------------------------------------------------


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2022-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import numpy as np
import matplotlib.pyplot as plt

class AutoMultiImagePlot(object):
  def __init__(self, min=None, max=None, title=None):
    self.rows = []
    self.row_count = 0
    self.row_titles = []
    self.current_row = -1
    self.row_col_count = dict()
    self.max_col_count = 0
    self.min = min
    self.max = max
    self.title = title

  def add_row(self, row_title=None):
    self.current_row = self.row_count
    self.rows.append([])
    self.row_count = len(self.rows)
    self.row_titles.append(row_title)

  def add_column(self, image, image_title=None, color_map=None, aspect=None, extent=None):
    oRowColumns = self.rows[self.current_row]
    dImage = {"image": image, "title": image_title
      , "cmap": color_map, "aspect": aspect
      , "extend": extent}

    oRowColumns.append(dImage)
    self.rows[self.current_row] = oRowColumns
    nColCount = len(oRowColumns)

    self.row_col_count[self.current_row] = nColCount
    if nColCount > self.max_col_count:
      self.max_col_count = nColCount

  def prepare(self, title=None, figure_size=(15, 6), restrict_columns=None):

    nColumns = restrict_columns
    if nColumns is None:
      nColumns = self.max_col_count
    fig, oSubplotGrid = plt.subplots(  nrows=self.row_count, ncols=nColumns
                                     , figsize=figure_size
                                     , subplot_kw={'xticks': [], 'yticks': []})
    bIsSingleRow = self.row_count == 1
    if bIsSingleRow:
      oSubplotGrid = oSubplotGrid[np.newaxis, ...]

    if title is None:
      title = self.title
    fig.suptitle(title)
    for nRowIndex, oRowColumns in enumerate(self.rows):
      if len(oRowColumns) > 0:
        sRowTitle = self.row_titles[nRowIndex]
        nRowImageCount = len(oRowColumns)
        #nIncr = nImageCount // nRowColumnCount
        nIncr = 1
        nImageIndex = 0
        for nColIndex in range(nColumns):
          bMustPlot = nColIndex < nRowImageCount
          #if (nIncr == 0) and (nColIndex > 0):
          #  bMustPlot = False

          if bMustPlot:
            dImage = oRowColumns[nImageIndex]
            oSubPlot = oSubplotGrid[nRowIndex, nColIndex]
            sTitle = dImage['title']
            if sTitle is not None:
              oSubPlot.title.set_text(sTitle)
            oSubPlot.set_xticks([])
            oSubPlot.set_yticks([])
            oSubPlot.imshow(dImage["image"], cmap=dImage["cmap"],
                            aspect=dImage["aspect"], extent=dImage["extend"],
                            vmin=self.min, vmax=self.max
                            )

            if nColIndex == 0:
              if sRowTitle is not None:
                oSubPlot.text(0.0, 0.5, sRowTitle, transform=oSubPlot.transAxes,
                              horizontalalignment='right', verticalalignment='center',
                              fontsize=9, fontweight='bold')
          nImageIndex += nIncr
    fig.subplots_adjust(wspace=0.1, hspace=0.6)
    return self

  # --------------------------------------------------------------------------------------
  def save(self, filename):
    plt.savefig(filename, bbox_inches='tight')
    return self

  # --------------------------------------------------------------------------------------
  def show(self):
    plt.show()
  # --------------------------------------------------------------------------------------



import matplotlib.pyplot as plt  # use the subpackage (a.k.a. namespace) with the alias "plt"
import numpy as np
from matplotlib import colors
from radnn.data import DataPreprocessingKind

# ====================================================================================================
class PlotDataset(object):  # class CPlot: object
  # --------------------------------------------------------------------------------------
  # Constructor
  def __init__(self, samples, labels
               , label_descriptions=["orange tree", "olive tree"]
               , colors=["darkorange", "darkseagreen"]
               , x_label="Feature 1"
               , y_label="Feature 2"
               ):
    # ................................................................
    # // Fields \\
    self.title = ""
    self.samples = samples
    self.labels = labels
    self.label_descriptions = label_descriptions
    self.colors = colors  # # https://matplotlib.org/3.1.0/gallery/color/named_colors.html
    self.x_label = x_label
    self.y_label = y_label
    # ................................................................

  # --------------------------------------------------------------------------------------
  def prepare(self, title, line_slope=None, line_intercept=None, limits_x=None, limits_y=None, preprocessed: DataPreprocessingKind | None = None):
    self.title = title

    # Two dimensional dataset for the scatter plot
    nXValues = self.samples[:, 0]
    nYValues = self.samples[:, 1]
    nLabels = self.labels

    oColorMap = colors.ListedColormap(self.colors)

    fig, ax = plt.subplots(figsize=(8, 8))
    plt.scatter(nXValues, nYValues, c=nLabels, cmap=oColorMap)

    plt.title(self.title)
    cb = plt.colorbar()
    nLoc = np.arange(0, max(nLabels), max(nLabels) / float(len(self.colors)))
    cb.set_ticks(nLoc)
    oLegend = [ f"{i}:{x}" for i,x in enumerate(self.label_descriptions)]
    cb.set_ticklabels(oLegend)

    if (line_slope is not None):
      x1 = np.min(nXValues)
      y1 = line_slope * x1 + line_intercept;
      x2 = np.max(nXValues)
      y2 = line_slope * x2 + line_intercept;
      oPlot1 = ax.plot([x1, x2], [y1, y2], 'r--', label="Decision line")
      oLegend = plt.legend(loc="upper left", shadow=True, fontsize='x-large')
      oLegend.get_frame().set_facecolor("lightyellow")

    if preprocessed == DataPreprocessingKind.MIN_MAX_NORMALIZE:
      ax.set_xlim((-0.05, 1.05))
      ax.set_ylim((-0.05, 1.05))
    if preprocessed == DataPreprocessingKind.STANDARDIZE:
      ax.set_xlim((-4.05, 4.05))
      ax.set_ylim((-4.05, 4.05))
    else:
      if limits_x is not None:
        ax.set_xlim(limits_x[0], limits_x[1])
      if limits_y is not None:
        ax.set_ylim(limits_y[0], limits_y[1])

    ax.set_xlabel(self.x_label)
    ax.set_ylabel(self.y_label)

    # plt.scatter(oDataset.Samples[:,0], oDataset.Samples[:,1])
    # , t, 'r--', t, t**2, 'bs', t, t**3, 'g^')
    return self
  # --------------------------------------------------------------------------------------
  def save(self, filename):
    plt.savefig(filename, bbox_inches='tight')
    return self
  # --------------------------------------------------------------------------------------
  def show(self):
    plt.show()
  # --------------------------------------------------------------------------------------


# ====================================================================================================

class CPlot(PlotDataset):
  # --------------------------------------------------------------------------------------
  def __init__(self, p_sTitle, p_nSamples, p_nLabels, p_sLabelDescriptions=["orange tree", "olive tree"]
               , p_sColors=["darkorange", "darkseagreen"]
               , p_sXLabel="Feature 1"
               , p_sYLabel="Feature 2"
               ):
    super().__init__(p_nSamples, p_nLabels, p_sLabelDescriptions, p_sColors, p_sXLabel, p_sYLabel)
    self.s_title = p_sTitle

  # --------------------------------------------------------------------------------------
  def Show(self, p_bIsMinMaxScaled=False, p_nLineSlope=None, p_nLineIntercept=None, p_nLimitsX=[-4, 4], p_nLimitsY=[-4, 4]):
    if p_bIsMinMaxScaled:
      self.prepare(self.s_title, p_nLineSlope, p_nLineIntercept, p_nLimitsX, p_nLimitsY, preprocessed=DataPreprocessingKind.MIN_MAX_NORMALIZE).show()
    else:
      self.prepare(self.s_title,  p_nLineSlope, p_nLineIntercept, p_nLimitsX, p_nLimitsY).show()
  # --------------------------------------------------------------------------------------



# ......................................................................................
# MIT License

# Copyright (c) 2019-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
import os
import numpy as np
import pandas as pd
from abc import ABC, abstractmethod
from .sample_set_simple import SampleSet
from .sample_set_kind import SampleSetKind
from .sample_preprocessor import SamplePreprocessor, VoidPreprocessor
from .errors import *
from radnn import FileStore
from radnn import mlsys
from .constants import DataPreprocessingKind
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler

# ======================================================================================================================
class DataSetCallbacks(object):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, download_method=None, seed_method=None):
    self._lazy_download = download_method
    self._random_seed = seed_method
  # --------------------------------------------------------------------------------------------------------------------
  def lazy_download(self, fs):
    self._lazy_download(fs)
  # --------------------------------------------------------------------------------------------------------------------
  def random_seed(self, seed: int):
    self._random_seed(seed)
  # --------------------------------------------------------------------------------------------------------------------
# ======================================================================================================================s





# ======================================================================================================================
class DataSetBase(ABC):
  # --------------------------------------------------------------------------------------------------------------------
  # Constructor
  def __init__(self, name: str, variant: str | None = None, file_store=None, random_seed: int | None = None,
               callbacks: DataSetCallbacks | None = None):
    # ..................// Instance Fields \\.........................
    self.name = name
    self.variant = variant
    self.fs = None
    self._determine_local_filestore(file_store)
    assert self.fs is not None, ERR_DS_MUST_PROVIDE_LOCAL_FILESTORE
    self.random_seed = random_seed
    
    self.all_ids = None
    self.all_samples = None
    self.all_labels = None
    
    self.feature_count = None
    self.class_count = None
    self.sample_count = None

    self.callbacks: DataSetCallbacks = callbacks
    
    self.hprm: dict | None = None
    self.ts: SampleSet | None = None
    self.vs: SampleSet | None = None
    self.us: SampleSet | None = None
    self.preprocessor: SamplePreprocessor = VoidPreprocessor(self)
    
    self.is_split = False
    # ................................................................

  # --------------------------------------------------------------------------------------
  def split(self, validation_samples_pc=0.10,
            preprocessing: DataPreprocessingKind | None = DataPreprocessingKind.STANDARDIZE,
            random_seed: int=2021):
    
    nTS_Samples, nVS_Samples, nTS_Labels, nVS_Labels = train_test_split(self.all_samples, self.all_labels,
                                                                        test_size=validation_samples_pc,
                                                                        random_state=random_seed)
    if preprocessing == DataPreprocessingKind.MIN_MAX_NORMALIZE:
      self.preprocessor = MinMaxScaler().fit(nTS_Samples)
    elif preprocessing == DataPreprocessingKind.STANDARDIZE:
      self.preprocessor = StandardScaler().fit(nTS_Samples)
    else:
      self.preprocessor = None

    if self.preprocessor is not None:
      nTS_Samples = self.preprocessor.transform(nTS_Samples)
      nVS_Samples = self.preprocessor.transform(nVS_Samples)

    # (Re)creating the subsets of the dataset after the splits have been created
    self.ts = SampleSet(self, nTS_Samples, nTS_Labels, kind=SampleSetKind.TRAINING_SET)
    if preprocessing == DataPreprocessingKind.STANDARDIZE:
      self.ts.mean = self.preprocessor.mean_
      self.ts.std = self.preprocessor.scale_

    self.vs = SampleSet(self, nVS_Samples, nVS_Labels, kind=SampleSetKind.VALIDATION_SET)
    
    self.is_split = True
    return self
  # --------------------------------------------------------------------------------------------------------------------
  @property
  def dataset_code(self):
    sUniqueName = f"{self.name.upper()}"
    if self.variant is not None:
      sUniqueName += f"_{self.variant.upper()}"
    return sUniqueName
  # --------------------------------------------------------------------------------------------------------------------
  def _determine_local_filestore(self, file_store):
    if (file_store is not None):
      if isinstance(file_store, FileStore):
        self.fs = file_store
      elif isinstance(file_store, str):
        if not os.path.exists(file_store):
          raise Exception(ERR_DS_FOLDER_NOT_FOUND % file_store)
        self.fs = FileStore(file_store)
    else:
      assert mlsys.filesys is not None, ERR_MLSYS_FILESYS_NOT_INITIALIZED
      
      self.fs: FileStore = mlsys.filesys.datasets.subfs(self.dataset_code)
  # --------------------------------------------------------------------------------------------------------------------
  @property
  def filesystem_folder(self):
    return self.fs.absolute_path
  # --------------------------------------------------------------------------------------------------------------------
  def read_hyperparams(self):
    pass # Optionally override
  # --------------------------------------------------------------------------------------------------------------------
  @abstractmethod
  def load_data(self):
    pass # Must implement
  # --------------------------------------------------------------------------------------------------------------------
  def load_cache(self, is_vector_samples=True, is_last_axis_features=True):
    nSuffix = ""
    if is_vector_samples:
      nSuffix = "-vec"
    elif not is_last_axis_features:
      nSuffix = "-torch"

    nTSSamples = self.fs.obj.load(f"{self.dataset_code}-TS-Samples{nSuffix}.pkl")
    nVSSamples = self.fs.obj.load(f"{self.dataset_code}-VS-Samples{nSuffix}.pkl")

    nTSLabels = self.fs.obj.load(f"{self.dataset_code}-TS-Labels{nSuffix}.pkl")
    nVSLabels = self.fs.obj.load(f"{self.dataset_code}-VS-Labels{nSuffix}.pkl")

    return nTSSamples, nVSSamples, nTSLabels, nVSLabels
  # --------------------------------------------------------------------------------------------------------------------
  def save_cache(self, ts_samples, vs_samples, ts_labels, vs_labels, is_vector_samples=True, is_last_axis_features=True):
    nSuffix = ""
    if is_vector_samples:
      nSuffix = "-vec"
    elif not is_last_axis_features:
      nSuffix = "-torch"
    self.fs.obj.save(ts_samples, f"{self.dataset_code}-TS-Samples{nSuffix}.pkl")
    self.fs.obj.save(vs_samples, f"{self.dataset_code}-VS-Samples{nSuffix}.pkl")

    self.fs.obj.save(ts_labels, f"{self.dataset_code}-TS-Labels{nSuffix}.pkl")
    self.fs.obj.save(vs_labels, f"{self.dataset_code}-VS-Labels{nSuffix}.pkl")
  # --------------------------------------------------------------------------------------------------------------------
  def prepare(self, hyperparams: dict | None = None):
    self.hprm = hyperparams
    
    # VIRTUAL CALL: Reads the hyperparameters into instance variables
    if self.hprm is not None:
      self.read_hyperparams()
    
    if (self.callbacks is not None):
      assert self.callbacks._lazy_download is not None, ERR_DS_NO_RANDOM_SEED_INITIALIZER_CALLBACK
      if self.callbacks._lazy_download is not None:
        self.callbacks.lazy_download(self.fs)
    
    if (self.random_seed is not None):
      bIsInitRandomSeed = False
      if self.callbacks is not None:
        if self.callbacks._random_seed is not None:
          self.callbacks.random_seed(self.random_seed)
          bIsInitRandomSeed = True
      if not bIsInitRandomSeed:
        mlsys.random_seed_all(self.random_seed)

    self.ts = None
    self.vs = None
    self.us = None

    # VIRTUAL CALL:  Imports the dataset from the source local/remote filestore to the local cache.
    self.load_data()
    
    if self.is_split:
      assert self.ts is not None, ERR_DS_SUBSET_MUST_HAVE_TS
      assert self.ts.kind == SampleSetKind.TRAINING_SET, ERR_DS_SUBSET_INVALID_SETUP
      if self.vs is not None:
        assert self.vs.kind == SampleSetKind.VALIDATION_SET, ERR_DS_SUBSET_INVALID_SETUP
        
    if self.us is not None:
      assert self.us.kind == SampleSetKind.UNKNOWN_TEST_SET, ERR_DS_SUBSET_INVALID_SETUP
      
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def assign(self, data, label_columns: range):
    self.all_samples, self.all_labels, self.all_ids = None, None, None
    if isinstance(data, tuple):
      self.all_samples, self.all_labels = data
    elif isinstance(data, np.ndarray):
      self.all_samples = data
    elif isinstance(data, dict):
      if ("samples" in dict):
        self.all_samples = data["samples"]
      if ("labels" in dict):
        self.all_labels = data["labels"]
      if ("ids" in dict):
        self.all_ids = data["ids"]
    elif isinstance(data, pd.DataFrame):
      if isinstance(data.columns, pd.Index):
        nData = data.iloc[1:].to_numpy()
      else:
        nData = data.to_numpy()
      
      if label_columns is None:
        self.all_samples = nData
      else:
        if label_columns.start >= 0:
          if label_columns.stop is None:
            self.all_labels = nData[:, label_columns.start]
            self.all_samples = nData[:, label_columns.start + 1:]
          else:
            self.all_labels = nData[:, label_columns.start:label_columns.stop + 1]
            self.all_samples = nData[:, label_columns.stop + 1:]
        else:
          self.all_samples = nData[:, :label_columns.start]
          self.all_labels = nData[:, label_columns.start:]
      
    if self.all_ids is None:
      self.all_ids = range(len(self.all_samples)) + 1
          
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def print_info(self):
    print(f"Dataset [{self.dataset_code}]")
    self.ts.print_info()
    if self.vs is not None:
      self.vs.print_info()
    if self.us is not None:
      self.us.print_info()
  # --------------------------------------------------------------------------------------------------------------------



# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
from typing import Dict, Type, Any

from radnn.errors import *
from .data_hyperparams import DataHyperparams
from radnn import mlsys



# ======================================================================================================================
class RadnnDataset(type):
  def __new__(mcls, name: str, bases: tuple[type, ...], ns: dict[str, Any]):
    cls = super().__new__(mcls, name, bases, ns)
    # Avoid registering the abstract base itself
    if name != "DatasetBuildAdapter":
      key = ns.get("NAME", name)  # optional override via class attribute NAME
      if key in DatasetFactory.registry:
        raise KeyError(f"The dataset '{key}' has been already registered ")
      DatasetFactory.registry[key] = cls
      
      alt_keys = cls.name_alias()
      for alt_key in alt_keys:
        alt_key = alt_key + "BuildAdapter"
        if alt_key in DatasetFactory.registry:
          raise KeyError(f"The dataset '{alt_key}' has been already registered ")
        DatasetFactory.registry[alt_key] = cls
        
    return cls
# ======================================================================================================================



# ======================================================================================================================
class DatasetBuildAdapter(metaclass=RadnnDataset):
  # --------------------------------------------------------------------------------------------------------------------
  @classmethod
  def name_alias(self):
    return []
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, **kwargs):
    self.builder_kwargs = kwargs
    self.hprm: dict | None = self.builder_kwargs.get("hprm", None)
    self.hyperparams: DataHyperparams | None = self.builder_kwargs.get("model_hyperparams", None)
    if self.hyperparams is not None:
      self.hprm = self.hyperparams.hprm
    
    self.fs = self.builder_kwargs.get("dataset_fs", None)
    self.dataset = None
  # --------------------------------------------------------------------------------------------------------------------
  def build(self, **kwargs: Any):
    raise NotImplementedError
  # --------------------------------------------------------------------------------------------------------------------
# ======================================================================================================================







# ======================================================================================================================
class DatasetFactory(object):
  registry: Dict[str, Type["ModelBuildAdapter"]] = {}
  
  @classmethod
  def produce(cls, hyperparams: dict, **kwargs: Any) -> DatasetBuildAdapter:
    oDataset, oDatasetBuilder = None, None
    oDatasetHyperparams: DataHyperparams = DataHyperparams.from_dict(hyperparams)
    assert oDatasetHyperparams.sample_dims is not None, HPARAMS_DATA_SAMPLE_DIMS
    assert len(oDatasetHyperparams.sample_dims) > 0, HPARAMS_DATA_SAMPLE_DIMS
    
    sModelAdapterKey = oDatasetHyperparams.dataset_name + "BuildAdapter"
    if sModelAdapterKey in cls.registry:
      oDatasetFS = mlsys.filesys.datasets.subfs(oDatasetHyperparams.dataset_fs_name)
      
      cBuilder: DatasetBuildAdapter = DatasetFactory.registry[sModelAdapterKey]
      oDatasetBuilder = cBuilder(model_hyperparams=oDatasetHyperparams, dataset_fs=oDatasetFS, **kwargs)
      oDataset = oDatasetBuilder.build(hprm=hyperparams, **kwargs)
    
    if oDataset is None:
      raise Exception(MODEL_BUILDER_NOT_REGISTERED % oDatasetHyperparams.base_name)
    
    return oDatasetBuilder
# ======================================================================================================================



import os

from radnn.core import RequiredLibs
oReqs = RequiredLibs()
if oReqs.is_tensorflow_installed:
  import tensorflow.keras as ker

class KBestStateSaver(object):
  # -----------------------------------------------------------------------------------
  def __init__(self, experiment_fs, metric, verbose=2):
    self.experiment_fs = experiment_fs
    self.metric = metric
    self.checkpoint_fs = self.experiment_fs.subfs("checkpoints")

    sCheckPointPathTemplate = os.path.join(self.checkpoint_fs.base_folder, "{epoch}")

    self.Callback = ker.callbacks.ModelCheckpoint(  filepath=sCheckPointPathTemplate,
                                                    verbose=verbose, save_weights_only=True,
                                                    monitor="val_" + self.metric,
                                                    mode="max", save_best_only=True)
  # -----------------------------------------------------------------------------------


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
import zipfile
from .fileobject import FileObject

class ZipFile(FileObject):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None, is_verbose=False):
    super().__init__(filename, parent_folder, error_template)
    self.is_verbose = is_verbose
    self._opened_filename = None
    self._encoding = None
  
  # --------------------------------------------------------------------------------------------------------
  def save(self, source_path, must_replace=True):
    """
    Compress a folder (including subfolders) into a .zip file.

    Args:
        output_zip_path (str): The path where the .zip file will be saved.
    """
    dest_zip_file = os.path.join(self.parent_folder, self.filename)
    if must_replace:
      if os.path.exists(dest_zip_file):
        os.remove(dest_zip_file)
      
      # Create the zip file
      with zipfile.ZipFile(dest_zip_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for root, dirs, files in os.walk(source_path):
          for file in files:
            abs_path = os.path.join(root, file)
            # Get relative path for proper folder structure inside zip
            rel_path = os.path.relpath(abs_path, start=source_path)
            zipf.write(abs_path, rel_path)
  
  # --------------------------------------------------------------------------------------------------------
  def extract(self, dest_path, has_progress_bar=True, is_verbose=True):
    """
    Extracts a .zip file into the specified folder.

    Args:
        zip_filename (str): Path to the .zip file.
        destination_folder (str): Folder where the contents should be extracted.
    """
    source_zip_file = os.path.join(self.parent_folder, self.filename)
    
    # Ensure the zip file exists
    if not os.path.isfile(source_zip_file):
      raise FileNotFoundError(f"Zip file not found: {source_zip_file}")
    
    # Ensure destination directory exists
    os.makedirs(dest_path, exist_ok=True)
    
    # Extract the zip file
    with zipfile.ZipFile(source_zip_file, 'r') as zip_ref:
      file_list = zip_ref.infolist()
      total_files = len(file_list)
      
      oIterator = file_list
      if is_verbose and has_progress_bar:
        from tqdm import tqdm
        oIterator = tqdm(file_list, total=total_files, desc="Extracting", unit="file")
      
      for file_info in oIterator:
        zip_ref.extract(file_info, dest_path)
    if is_verbose:
      print(f" |_ Extracted '{source_zip_file}' to '{dest_path}'")
  # ----------------------------------------------------------------------------------



# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
import random
import numpy as np
import json
from .core import AIGridInfo, RequiredLibs
from .utils import classproperty, subdict_kwargs
from radnn.system import FileSystem

class MLSystem(object):
  _instance = None
  
  @classproperty
  def instance(cls):
    if cls._instance is None:
      cls._instance = cls()
    return cls._instance

  # --------------------------------------------------------------------------------------
  def __init__(self):
    self._is_random_seed_initialized = False
    self._filesys = None
    self._seed = None
    self.switches = dict()
    self.switches["IsDebuggable"] = False
    self.req_libs: RequiredLibs = RequiredLibs()
    self.initialize_system()
    
    self.framework = "other"
    self.device = "CPU"
    
    # Ensure cuBLAS reproducibility for torch and/or tensorflow
    if self.req_libs.is_torch_installed:
      os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"  # https://docs.nvidia.com/cuda/cublas/index.html#cublasApi_reproducibility
      import torch
    if self.req_libs.is_tensorflow_installed:
      os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
      import tensorflow as tf
      
    
    if self.req_libs.is_tensorflow_installed:
      self.framework = "tensorflow"
    # By priority use torch for model trainers and dataset iterators (overrides co-existing tensorflow)
    if self.req_libs.is_torch_installed:
      self.framework = "torch"
    
    self._info = None

    # Initialize default device
    if self.framework == "torch":
      self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    elif self.framework == "tensorflow":
      gpus = tf.config.list_physical_devices("GPU")
      if gpus:
        tf.config.set_visible_devices(gpus[0], "GPU")
        tf.config.experimental.set_memory_growth(gpus[0], True)
        self.device = "/GPU:0"
      else:
        self.device = "/CPU:0"
  # --------------------------------------------------------------------------------------
  @property
  def info(self):
    if self._info is None:
      self._info = AIGridInfo()
      self.info.discover_devices(self.framework)
    return self._info
  # --------------------------------------------------------------------------------------
  @property
  def filesys(self) -> FileSystem:
    return self._filesys

  @filesys.setter
  def filesys(self, value):
    self._filesys = value
  # --------------------------------------------------------------------------------------
  @property
  def seed(self):
    return self._seed
  # --------------------------------------------------------------------------------------
  def initialize_system(self, sysconfig_filename = None) -> dict:
    # ...........................................................
    def load_json(filename=None, encoding=None):
      dResult = None
      if filename is not None:
        if os.path.exists(filename):
          
          sJSON = None
          if os.path.isfile(filename):
            oEncodingToTry = ["utf-8", "utf-16", "latin1", "ascii"]  # Add more if needed
            if encoding is None:
              bIsLoaded = False
              for sEnc in oEncodingToTry:
                try:
                  with open(filename, "r", encoding=sEnc) as oFile:
                    sJSON = oFile.read()
                  bIsLoaded = True
                  break
                except (UnicodeDecodeError, UnicodeError):
                  continue
              if not bIsLoaded:
                raise ValueError("Unsupported encoding")
            else:
              with open(filename, "r", encoding=encoding) as oFile:
                sJSON = oFile.read()
          if sJSON is not None:
            dResult = json.loads(sJSON)
        
      return dResult
    # ...........................................................
    
    sFile = sysconfig_filename
    if sFile is None:
      sPath = os.environ.get("RADNN_CONFIG_FILE", None)
      if sPath is None:
        sPath = os.path.abspath(".")
      sFile = os.path.join(sPath, "mlsystem.json")
      
    
    dConfig = load_json(sFile)
    if dConfig is not None:
      dFileSystemConfig = subdict_kwargs(dConfig, "FileSystem")
      self._filesys = FileSystem(**dFileSystemConfig)
  # --------------------------------------------------------------------------------------
  def random_seed_all(self, seed, is_done_once=False, is_parallel_deterministic=False):
    '''
    We are seeding the number generators to get some amount of determinism for the whole ML training process.
    For Tensorflow it is not ensuring 100% deterministic reproduction of an experiment that runs on the GPU.
    
    :param seed:
    :param is_done_once:
    :param is_parallel_deterministic:
    :return:
    '''
    self._seed = seed
    
    bContinue = True
    if is_done_once:
      bContinue = (not self._is_random_seed_initialized)
    
    if bContinue:
      random.seed(seed)
      os.environ['PYTHONHASHSEED'] = str(seed)
      np.random.seed(seed)
      if self.req_libs.is_tensorflow_installed:
        import tensorflow as tf
        tf.compat.v1.reset_default_graph()
        if is_parallel_deterministic:
          tf.config.experimental.enable_op_determinism()  # Enable determinism for num_parallel_calls
        tf.random.set_seed(seed)
        tf.keras.utils.set_random_seed(seed)
      if self.req_libs.is_torch_installed:
        import torch
        torch.manual_seed(seed)
        # GPU and multi-GPU
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        # For GPU determinism
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        torch.use_deterministic_algorithms(True)
        
      self._is_random_seed_initialized = True
      print("(>) Random seed set to %d" % seed)
  # --------------------------------------------------------------------------------------


mlsys: MLSystem = MLSystem.instance







# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
import json
import glob
from collections import OrderedDict

from .fileobject import FileObject

#TODO: jsonpickle
#https://stackoverflow.com/questions/3768895/how-to-make-a-class-json-serializable

class JSONFile(FileObject):
  # ----------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None):
    super(JSONFile, self).__init__(filename, parent_folder, error_template, "json")
  # ----------------------------------------------------------------------------------
  def load(self, filename=None, encoding=None):
    filename = self._useFileName(filename)

    dResult = None
    if os.path.exists(filename):

      sJSON = None
      if os.path.isfile(filename):
        oEncodingToTry = ["utf-8", "utf-16", "latin1", "ascii"]  # Add more if needed
        if encoding is None:
          bIsLoaded = False
          for sEnc in oEncodingToTry:
            try:
              with open(filename, "r", encoding=sEnc) as oFile:
                sJSON = oFile.read()
              bIsLoaded = True
              break
            except (UnicodeDecodeError, UnicodeError):
              continue
          if not bIsLoaded:
            raise ValueError("Unsupported encoding")
        else:
          with open(filename, "r", encoding=encoding) as oFile:
            sJSON = oFile.read()
      if sJSON is not None:
        dResult = json.loads(sJSON, object_pairs_hook=OrderedDict)
    else:
      if self.error_template is not None:
        raise Exception(self.error_template % filename)

    return dResult
  # ----------------------------------------------------------------------------------
  def save(self, obj, filename=None, is_sorted_keys=False, is_utf8=False):
    filename = self._useFileName(filename)

    if obj is not None:
      if isinstance(obj, dict):
        sJSON = json.dumps(obj, sort_keys=is_sorted_keys, indent=4, ensure_ascii=is_utf8)
      else:
        sJSON = json.dumps(obj, default=lambda o: obj.__dict__, sort_keys=is_sorted_keys, indent=4, ensure_ascii=is_utf8)

      sEncoding = "utf-8"
      if not is_utf8:
        sEncoding = "utf-16"

      with open(filename, "w", encoding=sEncoding) as oFile:
        oFile.write(sJSON)
        oFile.close()
  # ----------------------------------------------------------------------------------
  @property
  def files(self, is_full_path=True):
    oResult = []
    if (self.parent_folder is not None):
      oJSONFiles = glob.glob(os.path.join(self.parent_folder, '*.json'))
      oJSONFiles = sorted(oJSONFiles, key=os.path.getmtime)

      for sJSONFile in oJSONFiles:
        if is_full_path:
          oResult.append(os.path.join(self.parent_folder, sJSONFile))
        else:
          oResult.append(sJSONFile)
    return oJSONFiles
  # ----------------------------------------------------------------------------------


# ......................................................................................
# MIT License

# Copyright (c) 2019-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................

# This implementation is ported from the Java/C# trees that have been developed by me
# for the lesson CS215 "Data Structures & Algorithms" at Anatolia American University.

import numpy as np

from typing import Type, Any, Callable, Optional, Iterable, Union, List


class TreeNodeList(list):
  """
  Python version of CTreeNodeList<T>:
  - inherits from list
  - enforces uniqueness and max_branching_factor in append_node
  """
  
  def __init__(self, max_branching_factor: int = np.inf):
    super().__init__()
    self._max_branching_factor = max_branching_factor
  
  @property
  def item_count(self) -> int:
    return len(self)
  
  @property
  def max_branching_factor(self) -> int:
    return self._max_branching_factor
  
  @max_branching_factor.setter
  def max_branching_factor(self, value: int) -> None:
    self._max_branching_factor = int(value)
  
  def contains(self, node: Type["TreeNode"]) -> bool:
    return node in self
  
  def append_node(self, node: Type["TreeNode"]) -> None:
    if node is None:
      return
    if len(self) >= self._max_branching_factor:
      return
    if node not in self:
      super().append(node)
  
  def remove_node(self, node: Type["TreeNode"]) -> None:
    if node in self:
      super().remove(node)
  
  def __str__(self) -> str:
    lines = []
    for n in self:
      # C# used "[{Value}]".PadRight(16) + " " + Path
      v = getattr(n, "value", None)
      left = f"[{v}]"
      left = left + (" " * max(0, 16 - len(left)))
      lines.append(f"{left} {n.path}")
    return "\r\n".join(lines)


class TreeNode(object):
  def __init__(self):
    
    self.name: str = ""
    self.index: int = -1
    self.value: Any = None
    self._parent: TreeNode | None = None
    self._children: TreeNodeList = TreeNodeList()
    
  # -------------------------
  # Properties (C#-like)
  # -------------------------
  @property
  def children(self) -> TreeNodeList:
    return self._children
  
  @property
  def parent(self) -> Type["TreeNode"] | None:
    return self._parent
  
  @parent.setter
  def parent(self, new_parent: Type["TreeNode"] | None):
    # Remove from current parent
    if self._parent is not None:
      self._parent.children.remove_node(self)
    
    self._parent = new_parent
    
    # Add to new parent
    if self._parent is not None:
      self._parent.children.append_node(self)
  
  @property
  def is_root(self) -> bool:
    return self._parent is None
  
  @property
  def is_leaf(self) -> bool:
    return self._children.item_count == 0
  
  @property
  def child_count(self) -> int:
    return len(self._children)
  
  def __getitem__(self, index: int) -> Type["TreeNode"] | None:
    try:
      return self._children[index]
    except IndexError:
      return None
  
  @property
  def level(self) -> int:
    if self._parent is None:
      return 0
    return self._parent.level + 1
  
  @property
  def root(self) -> Type["TreeNode"]:
    if self._parent is None:
      return self
    return self._parent.root
  
  @property
  def path(self) -> str:
    if self._parent is None:
      return "/"
    if self._parent.is_root:
      return self._parent.path + self.name
    return self._parent.path + "/" + self.name
  
  def new_child(self, node_name_or_id: Union[str, int, None] = None) -> Type["TreeNode"]:
    child = TreeNode()
    child.index = len(self._children) + 1
    if node_name_or_id is None:
      child.name = str(child.index)
    else:
      child.name = str(node_name_or_id)
    child.parent = self
    return child
  
  def add_child(self, child_node: Type["TreeNode"]) -> int:
    child_node.parent = self
    return self._children.item_count - 1
  
  def remove_child(self, child_or_name: Union[Type["TreeNode"], str]) -> None:
    if isinstance(child_or_name, TreeNode):
      child_or_name.parent = None
    else:
      name = str(child_or_name)
      for c in list(self.children):
        if c.name == name:
          c.parent = None
          break
    
  def delete(self) -> None:
    # Postorder delete: delete children first
    for c in list(self.children):
      c.delete()
    
    # Then remove self from parent
    if self._parent is not None:
      self._parent.remove_child(self)
  
  def __eq__(self, other: Type["TreeNode"]) -> bool:
    return self.name == other.name
  
  def __str__(self) -> str:
    return self.path




class TreeNodeQueue(list):
  @property
  def is_empty(self) -> bool:
    return len(self) == 0
  
  def enqueue(self, item: TreeNode) -> None:
    self.append(item)
  
  def peek(self) -> Optional[TreeNode]:
    return self[0] if len(self) == 0 else None
    
  def dequeue(self) -> Optional[TreeNode]:
    return self.pop[0] if len(self) == 0 else None



class Tree:
  def __init__(self, root: Optional[TreeNode] = None):
    self.root: TreeNode = root if root is not None else TreeNode()
    self._node_list: Optional[TreeNodeList] = None
    #self.comparison_by: Optional[Callable[[Any, Any], int]] = None  #TODO
  '''
  def compare(self, this_item: Any, other_item: Any) -> int:
    """
    Closest Python equivalent of C# compare(T,T):
    - If comparison_by provided, use it.
    - Else try normal Python comparisons.
    - If not comparable, return 1 (same default as C# code's nResult=1).
    """
    if self.comparison_by is not None:
      return int(self.comparison_by(this_item, other_item))
    
    try:
      if this_item == other_item:
        return 0
      # Python doesn't have CompareTo; approximate:
      return -1 if this_item < other_item else 1
    except Exception:
      return 1
  '''
  
  def clear(self) -> None:
    self.root.delete()
    self.root = TreeNode()
  
  def _recurse_preorder_append(self, current: TreeNode, depth: int) -> None:
    self._node_list.append_node(current)
    for child in current.children:
      self._recurse_preorder_append(child, depth + 1)
  
  def _recurse_postorder_append(self, current: TreeNode, depth: int) -> None:
    for child in current.children:
      self._recurse_postorder_append(child, depth + 1)
    self._node_list.append_node(current)
  
  def traverse_depth_first(self, is_preorder: bool = True) -> TreeNodeList:
    self._node_list = TreeNodeList()
    if is_preorder:
      self._recurse_preorder_append(self.root, 0)
    else:
      self._recurse_postorder_append(self.root, 0)
    return self._node_list
  
  def traverse_breadth_first(self) -> TreeNodeList:
    node_list = TreeNodeList()
    q = TreeNodeQueue()
    q.enqueue(self.root)
    
    while not q.is_empty:
      node = q.dequeue()
      if node is None:
        continue
      node_list.append_node(node)
      for child in node.children:
        q.enqueue(child)
    
    return node_list
  
  def _recurse_follow_path(self, path_names: list, current: TreeNode, depth: int) -> Optional[TreeNode]:
    next_name = path_names.pop(0)
    if next_name is None:
      return None
    
    for child in current.children:
      if child.name == next_name:
        if len(path_names) == 0:
          return child
        return self._recurse_follow_path(path_names, child, depth + 1)
    
    return None
  
  def follow(self, path: str) -> Optional[TreeNode]:
    # Split by '/'
    parts = path.split("/")
    
    q = list()
    for p in parts:
      q.append(p)
    
    # In an empty tree the result will be the root node
    result: Optional[TreeNode] = self.root
    if not q.is_empty:
      q.pop(0)  # remove "" representing root when path starts with "/"
      result = self._recurse_follow_path(q, self.root, 1)
    
    return result
  
  def _indent(self, depth: int) -> str:
    if depth - 1 >= 0:
      return " " * ((depth - 1) * 4)
    return ""
  
  def _recurse_node_description(self, current: Optional[TreeNode], depth: int) -> str:
    if current is None or not current.is_root:
      prefix = self._indent(depth) + "|__ "
    else:
      prefix = ">"
    
    if current is None:
      return prefix
    
    s = prefix + current.name
    for child in current.children:
      s += "\r\n" + self._recurse_node_description(child, depth + 1)
    return s
  
  def __str__(self) -> str:
    return self._recurse_node_description(self.root, 0)



# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2022-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import matplotlib.pyplot as plt



class PlotFunction(object):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, title=None):
    self.title = title
  # --------------------------------------------------------------------------------------------------------------------
  def prepare(self, x_series, y_series, title=None, linewidth=3.0):
    if title is not None:
      self.title = title
    plt.title(self.title)
    plt.plot(x_series, y_series, linewidth=linewidth)
    plt.xlabel("x")
    plt.ylabel("y")
    plt.axhline(
      y=0.0,  # y-value of the horizontal line
      linestyle='--',  # dashed line
      color='gray',
      linewidth=1.0
    )
    plt.axvline(
      x=0.0,
      linestyle='--',
      color='gray',
      linewidth=1.0
    )
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def save(self, filename):
    plt.savefig(filename, bbox_inches='tight')
    return self
  # --------------------------------------------------------------------------------------------------------------------
  def show(self):
    plt.show()
  # --------------------------------------------------------------------------------------------------------------------
  
  


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2026 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
import sys
import shutil

from .fileobject import FileObject
if (sys.version_info.major == 3) and (sys.version_info.minor <= 7):
  import pickle5 as pickle
else:
  import pickle

class PickleFile(FileObject):
  LOCAL_CACHE = None
  # ----------------------------------------------------------------------------------
  def __init__(self, filename, parent_folder=None, error_template=None, is_verbose=False):
    super(PickleFile, self).__init__(filename, parent_folder, error_template, "pkl")
    self.is_verbose = is_verbose
    self.has_local_cache = False #TODO: Create a cache for pickle files on a fast drive
    self._cached_filename = None
    self._cached = None

    if (self.has_local_cache) and (not os.path.exists(PickleFile.LOCAL_CACHE)):
      os.makedirs(PickleFile.LOCAL_CACHE)
  # ----------------------------------------------------------------------------------
  def get(self, data_url):
    sParts = data_url.split("/")
    filename = sParts[0]
    dataprop = sParts[1]
    if self._cached_filename is None:
      self._cached_filename = filename
      self._cached = self.load(filename)
    return self._cached[dataprop]
  # ----------------------------------------------------------------------------------
  def close(self):
    self._cached_filename = None
    del self._cached
    self._cached = None
  # ----------------------------------------------------------------------------------
  def load(self, filename: str|None = None, is_python2_format: bool=False, error_template: str|None=None):
    """
    Deserializes the dataset from a pickle file if it exists.
    Parameters
        p_sFileName        : Full path to the  python object file
    Returns
        The object with its dataset or None when the file is not found.
    """
    sFileName = self._useFileName(filename)
    oData = None

    if (self.has_local_cache):
      sOriginalFileName = sFileName
      sFileName = os.path.join(PickleFile.LOCAL_CACHE, sFileName.replace(os.path.sep, "_"))
      if os.path.isfile(sOriginalFileName):
        if not os.path.exists(sFileName):
          print("... Caching file [%s]" % sFileName)
          shutil.copyfile(sOriginalFileName, sFileName)

    if os.path.isfile(sFileName):
      if self.is_verbose:
        print("      {.} Loading dataset from %s" % sFileName)

      with open(sFileName, "rb") as oFile:
        if is_python2_format:
          oUnpickler = pickle._Unpickler(oFile)
          oUnpickler.encoding = 'latin1'
          oData = oUnpickler.load()
        else:
          oData = pickle.load(oFile)
        oFile.close()
    else:
      if error_template is not None:
        raise Exception(error_template % sFileName)

    return oData
  # ----------------------------------------------------------------------------------
  def save(self, obj, filename: str | None =None, is_overwriting: bool =False, extra_display_label: str|None=None):
    """
    Serializes the dataset to a pickle file if it does not exists.
    Parameters
        p_sFileName        : Full path to the  python object file
    Returns
        True if a new file was created
    """
    bResult = False
    sFileName = self._useFileName(filename)

    if is_overwriting:
      bMustContinue = True
    else:
      bMustContinue = not os.path.isfile(sFileName)

    if bMustContinue:
      if self.is_verbose:
        if extra_display_label is not None:
          print("  {%s} Saving dataset to %s" % (extra_display_label, sFileName))
        else:
          print("  {.} Saving dataset to %s" % sFileName)
      with open(sFileName, "wb") as oFile:
        pickle.dump(obj, oFile, pickle.HIGHEST_PROTOCOL)
        oFile.close()
      bResult = True
    else:
      if self.is_verbose:
        if extra_display_label is not None:
          print("  {%s} Not overwritting %s" % (extra_display_label, sFileName))
        else:
          print("  {.} Not overwritting %s" % sFileName)

    return bResult
  # ----------------------------------------------------------------------------------

import platform
if platform.system() == "Windows":
  PickleFile.LOCAL_CACHE = "C:\\MLCache"
else:
  PickleFile.LOCAL_CACHE = "/MLCache"





# ......................................................................................
# MIT License

# Copyright (c) 2013 Andy Hayden

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
# Based on the blog post:  https://andyhayden.com/2013/dotable-dictionaries
class Dictionary(dict):
  __getattr__ = dict.__getitem__

  def __init__(self, d):
    super().__init__()
    self.update(**dict((k, self.parse(v))
                       for k, v in d.iteritems()))

  @classmethod
  def parse(cls, v):
    if isinstance(v, dict):
      return cls(v)
    elif isinstance(v, list):
      return [cls.parse(i) for i in v]
    else:
      return v


class MLModelFreezer(object):
  def __init__(self, model, hyperparams):
    self.model = model
    self.hyperparams = hyperparams
    self.model_name = self.hyperparams["Model.Name"]
    self.variant = self.hyperparams.get("Training.FineTuning.Freeze", None)
    
    if self.model_name == "MobileNet_V3_Large_Weights.IMAGENET1K_V1":
      self.mobilenet_frozen()
    elif self.model_name == "resnet50_Weights.IMAGENET1K_V1":
      self.resnet50_frozen()
    elif self.model_name == "efficientnet_b0_Weights.IMAGENET1K_V1":
      self.efficientnet_frozen()
    elif self.model_name == "convnext_tiny_Weights.IMAGENET1K_V1":
      self.convnext_frozen()
    elif self.model_name == "beitv2_large_patch16_224.in1k_ft_in22k_in1k":
      self.beitv2_frozen_2()
      
  def freeze_batch_norm(self):
    oNamedParams = list(iter(self.model .named_parameters()))
    for nIndex, oTuple in enumerate(oNamedParams):
      name, p = oTuple
      if ".bn" in name:
        p.requires_grad = False
  
  def freeze_layer_norm(self):
    oNamedParams = list(iter(self.model .named_parameters()))
    for nIndex, oTuple in enumerate(oNamedParams):
      name, p = oTuple
      if (".norm" in name) or (".gamma" in name) or (".beta" in name) :
        p.requires_grad = False
    
  def freeze_block(self, module, exclude_normalization=False):
    oParams = list(iter(module.parameters()))
    oNamedParams = list(iter(module.named_parameters()))
    for nIndex, oTuple in enumerate(oNamedParams):
      name, p = oTuple
      p = oParams[nIndex]
      bContinue = True
      if exclude_normalization:
        bContinue = not ".bn" in name
      if bContinue:
        p.requires_grad = False
  
  def mobilenet_frozen(self):
    """
    torchvision.mobilenet_v3_large has:
      model.features: Sequential([...])
      model.classifier: Sequential([...])

    We'll freeze by coarse slices of `features`.
    The exact indices can differ slightly across torchvision versions, but this is stable enough:
      - features[0] is stem
      - features[1:-1] are inverted residual blocks
      - features[-1] is final 1x1 conv block before pooling/classifier
    """
    if self.variant is None:
      return

    if not hasattr(self.model, "features"):
      return

    feats = self.model.features
    n = len(feats)

    # Define slices (roughly early/mid/late)
    # Stem
    if 0 in self.variant:
      self.freeze_block(feats[0])

    # Early inverted residual blocks
    if 1 in self.variant:
      end = max(2, n // 4)
      self.freeze_block(feats[1:end])

    # Mid blocks
    if 2 in self.variant:
      start = max(1, n // 4)
      end = max(start + 1, n // 2)
      self.freeze_block(feats[start:end])

    # Late blocks (excluding last conv)
    if 3 in self.variant:
      start = max(1, n // 2)
      end = max(start + 1, n - 1)
      self.freeze_block(feats[start:end])


      
  
  def resnet50_frozen(self):
    if self.variant is None:
      return
    
    self.freeze_batch_norm()
    
    if 0 in self.variant:
      # Stem
      self.freeze_block(self.model.step_conv)
      self.freeze_block(self.model.bn1)
      
    if 1 in self.variant:
      # First block of multiple conv layers
      self.freeze_block(self.model.block1)

    if 2 in self.variant:
      self.freeze_block(self.model.block2)

    if 3 in self.variant:
      self.freeze_block(self.model.block3)

    if 4 in self.variant:
      # Freezing all layers will train only the classification head
      self.freeze_block(self.model.layer4)


  def efficientnet_frozen(self):
    """
    torchvision.efficientnet_b0 has:
      model.features: Sequential of stem + stages + head conv
      model.classifier

    Commonly len(features) == 9 in torchvision:
      0 stem
      1..7 MBConv/FusedMBConv stages
      8 head conv
    We'll freeze in coarse groups.
    """
    if self.variant is None:
      return

    if not hasattr(self.model, "features"):
      return

    feats = self.model.features
    n = len(feats)

    if n == 0:
      return

    # Stem
    if 0 in self.variant:
      self.freeze_block(feats[0])

    # Early stages
    if 1 in self.variant:
      # typically features[1:3]
      self.freeze_block(feats[1:min(3, n)])

    # Mid stages
    if 2 in self.variant:
      # typically features[3:5]
      self.freeze_block(feats[min(3, n):min(5, n)])

    # Late stages
    if 3 in self.variant:
      # typically features[5:7]
      self.freeze_block(feats[min(5, n):min(7, n)])



  def convnext_frozen(self):
    """
    torchvision.convnext_tiny has:
      model.features: Sequential of 8 nnmodules:
        0 stem
        1 stage1
        2 downsample1
        3 stage2
        4 downsample2
        5 stage3
        6 downsample3
        7 stage4
      model.classifier
    We'll freeze stem/stages (and their downsample ops together).
    """
    if self.variant is None:
      return

    if not hasattr(self.model, "features"):
      return

    feats = self.model.features
    n = len(feats)
    if n < 2:
      return

    # Stem
    if 0 in self.variant:
      self.freeze_block(feats[0])

    # Stage 1 (+ downsample after it if present)
    if 1 in self.variant:
      if n > 1: self.freeze_block(feats[1])
      if n > 2: self.freeze_block(feats[2])

    # Stage 2 (+ downsample)
    if 2 in self.variant:
      if n > 3: self.freeze_block(feats[3])
      if n > 4: self.freeze_block(feats[4])

    # Stage 3 (+ downsample)
    if 3 in self.variant:
      if n > 5: self.freeze_block(feats[5])
      if n > 6: self.freeze_block(feats[6])
  
  
  def beitv2_frozen(self):
    if self.variant is None:
      return
    
    freeze_patch_embed = 0 in self.variant
    # ----------------
    
    if freeze_patch_embed:
      self.freeze_block(self.model.patch_embed)

    blocks = self.model.blocks
    for i, block in enumerate(blocks):
      if i in self.variant:
        self.freeze_block(block)
    
  
  # -------------------------
  # BEiT v2 (timm or HuggingFace)
  # -------------------------
  def beitv2_frozen_2(self):
    """
    Supports:
      - timm-style: model.patch_embed, model.blocks (list/ModuleList), model.norm
      - HF-style:   model.beit.embeddings, model.beit.encoder.layer, model.beit.layernorm

    variant meaning (coarse, stage-like):
      0: freeze patch embedding (+ early norms if present)
      1: freeze first quarter of transformer blocks
      2: freeze first half
      3: freeze first 3/4
      4: freeze all blocks (train head only)

    If you prefer "freeze exact N blocks", you can pass an int in hyperparams instead of a list.
    """
    if self.variant is None:
      return

    # Allow variant to be an int: freeze first N blocks (and patch embed)
    if isinstance(self.variant, int):
      n_blocks_to_freeze = self.variant
      self._beit_freeze_first_n_blocks(n_blocks_to_freeze, freeze_patch_embed=True)
      return

    # Otherwise assume list-like stages
    vset = set(self.variant)

    self.freeze_layer_norm()
    
    # Try timm-style first
    if hasattr(self.model, "blocks"):
      blocks = self.model.blocks
      num_blocks = len(blocks)

      # 0: patch embedding / stem
      if 0 in vset:
        if hasattr(self.model, "patch_embed"):
          self.freeze_block(self.model.patch_embed)
        # some timm models have pos_embed / cls_token as params; those don't need freezing

      # Helper: freeze first k blocks
      def freeze_first_k(k):
        for b in blocks[:max(0, min(k, num_blocks))]:
          self.freeze_block(b)

      # Stage mapping
      if 1 in vset:
        freeze_first_k(max(1, num_blocks // 4))
      if 2 in vset:
        freeze_first_k(max(1, num_blocks // 2))
      if 3 in vset:
        freeze_first_k(max(1, (3 * num_blocks) // 4))
      

    # Fall back to HuggingFace-style
    elif hasattr(self.model, "beit"):
      beit = self.model.beit

      if 0 in vset and hasattr(beit, "embeddings"):
        self.freeze_block(beit.embeddings)

      layers = None
      if hasattr(beit, "encoder") and hasattr(beit.encoder, "layer"):
        layers = beit.encoder.layer

      if layers is not None:
        num_blocks = len(layers)

        def freeze_first_k(k):
          for b in layers[:max(0, min(k, num_blocks))]:
            self.freeze_block(b)

        if 1 in vset:
          freeze_first_k(max(1, num_blocks // 4))
        if 2 in vset:
          freeze_first_k(max(1, num_blocks // 2))
        if 3 in vset:
          freeze_first_k(max(1, (3 * num_blocks) // 4))

    

  def _beit_freeze_first_n_blocks(self, n_blocks_to_freeze: int, freeze_patch_embed: bool = True):
    # timm-style
    if hasattr(self.model, "blocks"):
      if freeze_patch_embed and hasattr(self.model, "patch_embed"):
        self.freeze_block(self.model.patch_embed)
      for b in self.model.blocks[:max(0, n_blocks_to_freeze)]:
        self.freeze_block(b)
      if hasattr(self.model, "norm") and n_blocks_to_freeze >= len(self.model.blocks):
        self.freeze_block(self.model.norm)
      return

    # HF-style
    if hasattr(self.model, "beit"):
      beit = self.model.beit
      if freeze_patch_embed and hasattr(beit, "embeddings"):
        self.freeze_block(beit.embeddings)
      if hasattr(beit, "encoder") and hasattr(beit.encoder, "layer"):
        for b in beit.encoder.layer[:max(0, n_blocks_to_freeze)]:
          self.freeze_block(b)
      if hasattr(beit, "layernorm") and hasattr(beit, "encoder") and hasattr(beit.encoder, "layer"):
        if n_blocks_to_freeze >= len(beit.encoder.layer):
          self.freeze_block(beit.layernorm)
  


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2019-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................

from radnn.core import RequiredLibs
oReqs = RequiredLibs()
if oReqs.is_tensorflow_installed:
  import tensorflow as tf
from radnn import mlsys
from radnn.data_beta.preprocess import Normalizer, Standardizer
from .data_feed import DataFeed

class TFClassificationDataFeed(DataFeed):
  # --------------------------------------------------------------------------------------------------------------------
  def do_random_crop(self, samples, labels):
    tPaddedImage = tf.image.pad_to_bounding_box(samples, self.padding_offset, self.padding_offset
                                                      , self.padding_target, self.padding_target)
    tResult = tf.image.random_crop(tPaddedImage, self.input_shape, seed=mlsys.seed)
    print("crop", tResult)
    return tResult, labels
  # --------------------------------------------------------------------------------------------------------------------
  def do_flip_left_right(self, samples, labels):
    tResult = tf.image.random_flip_left_right(samples, seed=mlsys.seed)
    print("flip", tResult)
    return tResult, labels
  # --------------------------------------------------------------------------------------------------------------------
  def build_augmentation(self, feed, augmentation_kind):
    if augmentation_kind == "random_crop":
      feed = feed.map(self.do_random_crop, num_parallel_calls=8)
    elif augmentation_kind == "random_flip_left_right":
      feed = feed.map(self.do_flip_left_right, num_parallel_calls=8)
    elif augmentation_kind == "random_cutout":
      pass # //TODO: Cutout
    return feed
  # --------------------------------------------------------------------------------------------------------------------
  def build_iterator(self):
    feed = None
    if self.subset_type.is_training_set:
      feed = tf.data.Dataset.from_tensor_slices((self.dataset.ts_samples, self.dataset.ts_labels))
    elif self.subset_type.is_validation_set:
      feed = tf.data.Dataset.from_tensor_slices((self.dataset.vs_samples, self.dataset.vs_labels))
    elif self.subset_type.is_unknown_test_set:
      feed = tf.data.Dataset.from_tensor_slices((self.dataset.ut_samples, self.dataset.ut_labels))
    return feed
  # --------------------------------------------------------------------------------------------------------------------
  def preprocess_normalize_onehot(self, samples, labels):
    tSamples = tf.cast(samples, tf.float32)
    tSamples = (tSamples - self.value_preprocessor.min) / (self.value_preprocessor.max - self.value_preprocessor.min)
    tTargetsOneHot = tf.one_hot(labels, self.dataset.class_count)
    return tSamples, tTargetsOneHot
  # --------------------------------------------------------------------------------------------------------------------
  def preprocess_standardize_onehot(self, samples, labels):
    tSamples = tf.cast(samples, tf.float32)
    tSamples = (tSamples - self.value_preprocessor.mean) / self.value_preprocessor.std
    tTargetsOneHot = tf.one_hot(labels, self.dataset.class_count)
    return tSamples, tTargetsOneHot
  # --------------------------------------------------------------------------------------------------------------------
  def preprocess_normalize(self, samples, labels):
    tSamples = tf.cast(samples, tf.float32)
    tSamples = (tSamples - self.value_preprocessor.min) / (self.value_preprocessor.max - self.value_preprocessor.min)
    return tSamples, labels
  # --------------------------------------------------------------------------------------------------------------------
  def preprocess_standardize(self, samples, labels):
    tSamples = tf.cast(samples, tf.float32)
    tSamples = (tSamples - self.value_preprocessor.mean) / self.value_preprocessor.std
    return tSamples, labels
  # --------------------------------------------------------------------------------------------------------------------
  def build_preprocessor(self, feed):
    if self._is_multiclass:
      if isinstance(self.value_preprocessor, Standardizer):
        feed = feed.map(self.preprocess_standardize_onehot, num_parallel_calls=8)
      elif isinstance(self.value_preprocessor, Normalizer):
        feed = feed.map(self.preprocess_normalize_onehot, num_parallel_calls=8)
    else:
      if isinstance(self.value_preprocessor, Standardizer):
        feed = feed.map(self.preprocess_standardize, num_parallel_calls=8)
      elif isinstance(self.value_preprocessor, Normalizer):
        feed = feed.map(self.preprocess_normalize, num_parallel_calls=8)
    return feed
  # --------------------------------------------------------------------------------------------------------------------
  def build_random_shuffler(self, feed):
    feed = feed.shuffle(self.sample_count_to_shuffle, seed=mlsys.seed)
    return feed
  # --------------------------------------------------------------------------------------------------------------------
  def build_minibatch_maker(self, feed):
    feed = feed.batch(self.batch_size)
    return feed
  # --------------------------------------------------------------------------------------------------------------------





# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
import os
from radnn.core import system_name
from radnn.system.filestore import FileStore
from radnn.system.files import JSONFile

# =======================================================================================================================
class FileSystem(object):
  # --------------------------------------------------------------------------------------------------------
  def __init__(self, config_folder="MLConfig", model_folder="MLModels", dataset_folder="MLData", must_exist=False, setup_filename="*auto*", is_custom_setup=True):
    '''
    Initializes the file system settings for an experiment
    :param config_folder: The folder that contains the experiment hyperparameter files.
    :param model_folder: The main folder where the models are stored.
    :param dataset_folder: The main folder under which dataset folder are stores.
    :param setup_filename: The filename of the setup file.
    :param must_exist: False: Auto-creates the directories on the file system | True: Raises an error if it does not exist.
    '''
    if config_folder is None:
      config_folder = "MLConfig"
    if model_folder is None:
      model_folder = "MLModels"
    if dataset_folder is None:
      dataset_folder = "MLData"

    
    self.setup_filename = setup_filename
    if self.setup_filename == "*auto*":
      self.setup_filename = system_name() + ".fsys"
    self._must_exist = must_exist
    self._has_loaded_groups = False
    self.model_groups = []

    dSetup = None
    if (self.setup_filename is not None) and (not is_custom_setup):
      if os.path.exists(self.setup_filename):
        oFile = JSONFile(self.setup_filename)
        dSetup = oFile.load()
        config_folder = dSetup["Configs"]
        model_folder = dSetup["Models"]
        dataset_folder = dSetup["Datasets"]
        if "ModelGroups" in dSetup:
          self.model_groups = dSetup["ModelGroups"]
          self._has_loaded_groups = True

    if dSetup is None:
      dSetup = dict()
      dSetup["Configs"] = config_folder
      dSetup["Models"] = model_folder
      dSetup["Datasets"] = dataset_folder


    if len(self.model_groups) > 0:
      for model_group in self.model_groups:
        config_folder   = os.path.join(config_folder, model_group)
        model_folder    = os.path.join(model_folder, model_group)

    # ...................... | Fields | ......................
    self.setup = dSetup
    self.configs: FileStore = FileStore(config_folder, must_exist=self._must_exist)
    self.models: FileStore = FileStore(model_folder, must_exist=self._must_exist)
    self.datasets: FileStore = FileStore(dataset_folder, must_exist=self._must_exist)
    # ........................................................
  # --------------------------------------------------------------------------------------------------------
  def group(self, group_name):
    if not self._has_loaded_groups:
      if group_name is not None:
          self.model_groups.append(group_name)
          self.setup["ModelGroups"] = self.model_groups

          self.configs   = self.configs.subfs(group_name, must_exist=self._must_exist)
          self.models    = self.models.subfs(group_name, must_exist=self._must_exist)

    return self
  # --------------------------------------------------------------------------------------------------------
  def save_setup(self):
    oFile = JSONFile(self.setup_filename)
    oFile.save(self.setup)
  # --------------------------------------------------------------------------------------------------------
  def __str__(self)->str:
    sResult = f"  FileSystem.ConfigFolder: \"{self.configs.base_folder}\",\n"
    sResult += f"  FileSystem.DatasetFolder: \"{self.datasets.base_folder}\"\n"
    sResult += f"  FileSystem.ModelFolder: \"{self.models.base_folder}\",\n"
    if len(self.model_groups) > 0:
      sModelGroups = str(self.model_groups)
      sModelGroups = sModelGroups.replace("'", "\"")
      sResult += f"  ModelGroups: {sModelGroups}\n"
    sResult = "{\n" + sResult + "}"
    return sResult
  # --------------------------------------------------------------------------------------------------------
  def __repr__(self)->str:
    return self.__str__()
  # --------------------------------------------------------------------------------------------------------
# =======================================================================================================================



# ......................................................................................
# MIT License

# Copyright (c) 2022-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................
import numpy as np

from .sample_set import SampleSet
from .dataset_base import DataSetBase, DataSetCallbacks


# ----------------------------------------------------------------------------------------------------------------------
def generate_sequence_clips(samples, labels, window_size, stride, is_padding_zeros=False):
  nSequenceIndex = 0
  while nSequenceIndex < samples.shape[0]:
    nLabel = labels[nSequenceIndex]
    nPosition = 0
    nSpanPoints = window_size
    if is_padding_zeros:
      nSpanPoints = window_size - 3 * stride

    nDataPointCount = samples.shape[1]
    while (nPosition + nSpanPoints) <= nDataPointCount:
      if is_padding_zeros and ((nPosition + window_size) >= nDataPointCount):
        nSeqSample = np.zeros((window_size, samples.shape[2]), np.float32)
        nSeqSample[nPosition + window_size - nDataPointCount:, :] = samples[nSequenceIndex, nPosition:, :]
      else:
        nSeqSample = samples[nSequenceIndex, nPosition:nPosition + window_size, :]

      yield (nSeqSample, nLabel)

      nPosition += stride
    nSequenceIndex += 1
# ----------------------------------------------------------------------------------------------------------------------





class SequenceDataSet(DataSetBase):
  # --------------------------------------------------------------------------------------------------------------------
  def __init__(self, name: str, variant: str|None=None, file_store=None, random_seed: int | None=None, callbacks: DataSetCallbacks | None = None):
    super().__init__(name, variant, file_store, random_seed, callbacks)

    #self, name,  fs, clip_window_size=None, clip_stride=None, is_padding_zeros=False, random_seed=None, is_classification=True):
    #super(SequenceDataSet, self).__init__(name, fs, random_seed, is_classification)

    self.clip_window_size: int|None = None
    self.clip_stride: int|None = None
    self.is_padding_zeros: bool|None = None
  # --------------------------------------------------------------------------------------------------------------------
  def read_hyperparams(self):
    self.clip_window_size = self.hprm.get("Data.Clip.WindowSize", None)
    self.clip_stride = self.hprm.get("Data.Clip.Stride", None)
    self.is_padding_zeros = self.hprm.get("Data.Clip.IsPaddingZeroes", None)
  # --------------------------------------------------------------------------------------------------------------------
  def generate_clips(self, subset: SampleSet):
    return generate_sequence_clips(self, subset.samples, subset.labels,
                                   self.clip_window_size, self.clip_stride, self.is_padding_zeros)
  # --------------------------------------------------------------------------------------------------------------------
  def convert_samples_to_clips(self, clip_window_size=None, clip_stride=None, is_padding_zeros=False):
    if clip_window_size is not None:
      self.clip_window_size = clip_window_size
    if clip_stride is not None:
      self.clip_stride = clip_stride
    if is_padding_zeros and (not self.is_padding_zeros):
      self.is_padding_zeros = is_padding_zeros

    # TODO: Reinstate dataset card
    '''
    self.card["clips.window_size"] = self.clip_window_size
    self.card["clips.stride"] = self.clip_stride
    self.card["clips.is_padding_zeros"] = self.is_padding_zeros
    '''

    # Create training set clips
    oSubsets = [self.ts, self.vs, self.us]
    nClips = []
    nClipLabels = []
    for (nClip, nClipLabel) in self.generate_clips():
      nClips.append(nClip)
      nClipLabels.append(nClipLabel)
    nClips = np.asarray(nClips)
    nClipLabels = np.asarray(nClipLabels)
    self.assign_training_set(nClips, nClipLabels)

    # Create validation set clips
    if self.vs_samples is not None:
      nClips = []
      nClipLabels = []
      for (nClip, nClipLabel) in self.vs_sequence_clips:
        nClips.append(nClip)
        nClipLabels.append(nClipLabel)
      nClips = np.asarray(nClips)
      nClipLabels = np.asarray(nClipLabels)
      self.assign_validation_set(nClips, nClipLabels)

    # Create unknown test set clips
    if self.ut_samples is not None:
      nClips = []
      nClipLabels = []
      for (nClip, nClipLabel) in self.ut_sequence_clips:
        nClips.append(nClip)
        nClipLabels.append(nClipLabel)
      nClips = np.asarray(nClips)
      nClipLabels = np.asarray(nClipLabels)
      self.assign_unknown_test_set(nClips, nClipLabels)

    return self
  # --------------------------------------------------------------------------------------


# ......................................................................................
# MIT License

# Copyright (c) 2023 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# ......................................................................................


import numpy as np
from sklearn.model_selection import train_test_split  # import a standalone procedure toyfunction from the pacakge


# =========================================================================================================================
class CDataSetBase(object):
  # --------------------------------------------------------------------------------------
  # Constructor
  def __init__(self, name=None, random_seed=None, is_classification=True):
    # ................................................................
    # // Fields \\
    self.Name = name
    self.RandomSeed = random_seed
    self.IsClassification = is_classification
    
    self.FeatureCount = None
    self.ClassCount = None
    self.class_names = dict()
    self.SampleCount = None
    
    self.Samples = None
    self.Targets = None
    
    self.TSSampleIDs = None
    self.TSSamples = None
    self.TSTargets = None
    self.TSSampleCount = 0
    
    self.VSSampleIDs = None
    self.VSSamples = None
    self.VSTargets = None
    self.VSSampleCount = 0
    
    self.UTSampleIDs = None
    self.UTSamples = None
    self.UTTargets = None
    self.UTSampleCount = None
    
    # ................................................................
    if self.RandomSeed is not None:
      ml.random_seed(self.RandomSeed)
  
  # --------------------------------------------------------------------------------------
  def preview_images(self):
    import matplotlib.pyplot as plt
    
    # Look at some sample images from dataset
    plt.figure(figsize=(10, 10))
    for i in range(25):
      plt.subplot(5, 5, i + 1)
      if self.TSSamples.shape[2] == 2:
        plt.imshow(self.TSSamples[i].squeeze().astype(np.uint8), cmap='gray')
      else:
        plt.imshow(self.TSSamples[i].squeeze().astype(np.uint8))
      
      nClassIndex = self.TSLabels[i]
      sClassDescr = str(nClassIndex)
      if isinstance(self.class_names, dict):
        if nClassIndex in self.class_names:
          sClassName = self.class_names[nClassIndex]
          sClassDescr += " `" + sClassName + "`"
      elif isinstance(self.class_names, list):
        sClassName = self.class_names[nClassIndex]
        sClassDescr += " `" + sClassName + "` "
      
      plt.title(f"Label: {sClassDescr}")
      plt.axis('off')
    plt.show()
  
  # --------------------------------------------------------------------------------------
  @property
  def Labels(self):
    return self.Targets
  
  # --------------------------------------------------------------------------------------
  @property
  def TSLabels(self):
    return self.TSTargets
  
  # --------------------------------------------------------------------------------------
  @property
  def VSLabels(self):
    return self.VSTargets
  
  # --------------------------------------------------------------------------------------
  @property
  def UTLabels(self):
    return self.UTTargets
  
  # --------------------------------------------------------------------------------------
  def TrainingSet(self, p_nSamples, p_nTargets):
    self.TSSamples = p_nSamples
    self.TSTargets = p_nTargets
    self.countSamples()
    self.TSSampleIDs = np.arange(0, self.TSSampleCount)
    
    # Feature count is calculated on samples that are flattened as vectors
    if self.FeatureCount is None:
      self.FeatureCount = np.prod(self.TSSamples.shape[1:])
    
    if self.ClassCount is None:
      if self.IsClassification:
        self.ClassCount = len(np.unique(self.TSTargets))
      else:
        self.ClassCount = 0
  
  # --------------------------------------------------------------------------------------
  def ValidationSet(self, p_nSamples, p_nTargets):
    self.VSSamples = p_nSamples
    self.VSTargets = p_nTargets
    self.countSamples()
    self.VSSampleIDs = np.arange(0, self.VSSampleCount)
  
  # --------------------------------------------------------------------------------------
  def UnknownTestSet(self, p_nSamples, p_nTargets):
    self.UTSamples = p_nSamples
    self.UTTargets = p_nTargets
    self.countSamples()
    self.UTSampleIDs = np.arange(0, self.UTSampleCount)
    # --------------------------------------------------------------------------------------
  
  def info(self):
    self.PrintInfo()
  
  # --------------------------------------------------------------------------------------
  def PrintInfo(self):
    print("Dataset [%s]" % self.Name)
    print("  |__ FeatureCount:", self.FeatureCount)
    if self.IsClassification:
      print("  |__ ClassCount:", self.ClassCount)
    
    if self.TSSamples is not None:
      print("  |__ Training set samples  : %d   shape:%s" % (self.TSSampleCount, self.TSSamples.shape))
    if self.TSTargets is not None:
      print("  |__ Training set targets  : %d   shape:%s" % (self.TSSampleCount, self.TSTargets.shape))
    
    if self.VSSamples is not None:
      print("  |__ Validation set samples: %d   shape:%s" % (self.VSSampleCount, self.VSSamples.shape))
    if self.VSTargets is not None:
      print("  |__ Validation set targets: %d   shape:%s" % (self.VSSampleCount, self.VSTargets.shape))
    
    if self.UTSamples is not None:
      print("  |__ MemoryTest set samples      : %d   shape:%s" % (self.UTSampleCount, self.UTSamples.shape))
    if self.UTTargets is not None:
      print("  |__ MemoryTest set targets      : %d   shape:%s" % (self.UTSampleCount, self.UTTargets.shape))
  
  # --------------------------------------------------------------------------------------
  def countSamples(self):
    if self.TSSamples is not None:
      self.TSSampleCount = self.TSSamples.shape[0]
      self.SampleCount = self.TSSampleCount + self.VSSampleCount
    
    if self.VSSamples is not None:
      self.VSSampleCount = self.VSSamples.shape[0]
      self.SampleCount = self.TSSampleCount + self.VSSampleCount
      
      # The test set samples are not included in the available sample count
    if self.UTSamples is not None:
      self.UTSampleCount = self.UTSamples.shape[0]
  
  # --------------------------------------------------------------------------------------
  def Split(self, p_nTrainingPercentage):
    nTSSamples, nVSSamples, nTSTargets, nVSTargets = train_test_split(self.Samples, self.Targets
                                                                      , test_size=1.0 - p_nTrainingPercentage
                                                                      , random_state=self.RandomSeed
                                                                      , shuffle=True
                                                                      , stratify=self.Targets
                                                                      )
    self.TrainingSet(nTSSamples, nTSTargets)
    self.ValidationSet(nVSSamples, nVSTargets)
    self.countSamples()
  
  # --------------------------------------------------------------------------------------
  def LoadCache(self, p_oFileStore, p_sSamplesFilePrefix="Samples", p_sTargetsFilePrefix="Targets", p_bIsVerbose=False):
    bResult = p_oFileStore.Exists("%s.pkl" % p_sSamplesFilePrefix) or p_oFileStore.Exists(
      "%s.TS.pkl" % p_sSamplesFilePrefix)
    
    if bResult:
      if p_bIsVerbose:
        print("Loading known dataset set ...")
      self.Samples = p_oFileStore.Deserialize("%s.pkl" % p_sSamplesFilePrefix)
      self.Targets = p_oFileStore.Deserialize("%s.pkl" % p_sTargetsFilePrefix)
      
      if p_bIsVerbose:
        print("Loading training set ...")
      nTSSamples = p_oFileStore.Deserialize("%s.TS.pkl" % p_sSamplesFilePrefix)
      nTSTargets = p_oFileStore.Deserialize("%s.TS.pkl" % p_sTargetsFilePrefix)
      self.TrainingSet(nTSSamples, nTSTargets)
      
      if p_bIsVerbose:
        print("Loading validation set ...")
      nVSSamples = p_oFileStore.Deserialize("%s.VS.pkl" % p_sSamplesFilePrefix)
      nVSTargets = p_oFileStore.Deserialize("%s.VS.pkl" % p_sTargetsFilePrefix)
      self.ValidationSet(nVSSamples, nVSTargets)
      
      if p_bIsVerbose:
        print("Loading unknown test dataset set ...")
      nUTSamples = p_oFileStore.Deserialize("%s.UT.pkl" % p_sSamplesFilePrefix)
      if nUTSamples is not None:
        nUTTargets = p_oFileStore.Deserialize("%s.UT.pkl" % p_sTargetsFilePrefix)
        self.UnknownTestSet(nUTSamples, nUTTargets)
    
    return bResult
  
  # --------------------------------------------------------------------------------------
  def SaveCache(self, p_oFileStore, p_sSamplesFilePrefix="Samples", p_sTargetsFilePrefix="Targets"):
    if self.Samples is not None:
      p_oFileStore.Serialize("%s.pkl" % p_sSamplesFilePrefix, self.Samples, True)
      p_oFileStore.Serialize("%s.pkl" % p_sTargetsFilePrefix, self.Labels, True)
    
    p_oFileStore.Serialize("%s.TS.pkl" % p_sSamplesFilePrefix, self.TSSamples, True)
    p_oFileStore.Serialize("%s.TS.pkl" % p_sTargetsFilePrefix, self.TSLabels, True)
    
    p_oFileStore.Serialize("%s.VS.pkl" % p_sSamplesFilePrefix, self.VSSamples, True)
    p_oFileStore.Serialize("%s.VS.pkl" % p_sTargetsFilePrefix, self.VSLabels, True)
    
    if self.UTSamples is not None:
      p_oFileStore.Serialize("%s.UT.pkl" % p_sSamplesFilePrefix, self.UTSamples, True)
      p_oFileStore.Serialize("%s.UT.pkl" % p_sTargetsFilePrefix, self.UTLabels, True)
  # --------------------------------------------------------------------------------------


# =========================================================================================================================


if __name__ == "__main__":
  ml.random_seed(2023)
  
  #nData = np.random.rand(100).astype(np.float32)
  nData = np.arange(0, 100)
  nData = nData / (nData + 1)
  print(nData.shape)
  nTargets = np.concatenate([np.zeros(50), np.ones(50)], axis=0).astype(np.int32)
  print(nTargets.shape)
  
  oDataset = CDataSetBase("dummy")
  oDataset.Samples = nData
  oDataset.Targets = nTargets
  oDataset.Split(0.8)
  oDataset.PrintInfo()
  
  nTSChecksum = np.sum(oDataset.TSSamples)
  nVSChecksum = np.sum(oDataset.VSSamples)
  print(nTSChecksum, nVSChecksum)
  
  # // Unit Testing \\
  assert nTSChecksum == 75.59110064555725
  assert nVSChecksum == 19.22152183680314
  
  # Seed None:                 75.45033306926038 19.362289413100005
  # Seed None: Parameter 2023: 75.59110064555725 19.22152183680314
  # Seed 2023: Parameter 2023: 75.59110064555725 19.22152183680314  # First call with the same seed
  # Seed 2023: Parameter None: 75.59110064555725 19.22152183680314  # First call with the same seed
  # Seed 2025:                 75.55118094349115 19.261441538869224



from .cnn.cnn_stem_setup import CNNSizeFactor, CNNStemSetup
from .model_hyperparams import ModelHyperparams
from .model_info import ModelInfo
from .model_factory import ModelFactory, ModelBuildAdapter




from .thread_context   import ThreadContext, ThreadEx
from .thread_worker   import ThreadWorker
from .semaphore_lock   import SemaphoreLock
from .thread_safe_queue  import ThreadSafeQueue
from .thread_safe_string_collection import ThreadSafeQueue



from .keras_best_state_saver import KBestStateSaver
from .keras_optimization_combo import KOptimizationCombo
from .keras_learning_rate_scheduler import KLearningRateScheduler
from .keras_learning_algorithm import KLearningAlgorithm


# ======================================================================================
#
#     Rapid Deep Neural Networks
#
#     Licensed under the MIT License
# ______________________________________________________________________________________
# ......................................................................................

# Copyright (c) 2018-2025 Pantelis I. Kaplanoglou

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# .......................................................................................
from time import sleep
from .thread_context import ThreadContext
from .thread_safe_queue import ThreadSafeQueue
from .thread_safe_string_collection import StringCollectionThreadSafe



class ThreadWorker(ThreadContext):
    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, name=None, p_nLoopSleepIntervalMS=100, p_oQueue=None, p_oLog=None, is_daemon_thread=True):
        super(ThreadWorker, self).__init__(name, is_daemon_thread=is_daemon_thread)
        #........................... |  Instance Attributes | ...........................
        self.SleepIntervalMsecs = p_nLoopSleepIntervalMS        
        self.SleepIntervalMsecs
        self.Queue = p_oQueue
        self.Log = None
        #................................................................................        

        # auto create the queue and its log
        if p_oQueue is None:
            self.Queue = ThreadSafeQueue()
        if p_oLog is None:
            self.Log = StringCollectionThreadSafe()
    # ------------------------------------------------------------------------------------------------------------------
    def ThreadMain(self, p_oArgs):
        nSleepInterval = float(self.SleepIntervalMsecs/1000)
        
        while self.must_continue:
            if not self.Queue.is_empty():
                oMessage = self.Queue.pop()
                if oMessage is not None:
                    self.ThreadInvokeMethod(oMessage)
            sleep(nSleepInterval)
    # ------------------------------------------------------------------------------------------------------------------
    def ThreadInvokeMethod(self, p_oMessage): #virtual
        pass
    # ------------------------------------------------------------------------------------------------------------------


